; *********************************************************************
; Vic 20 3Com 3C509B-TPO Ethernet card driver
; developed from the 3Com 3C509B-TPO Ethernet card driver in EhBASIC

; supports ARP, ICMP echo, passive TCP/IP and basic HTTP

; significant changes

; 04/02/29	-	first public release - V 00.44
; 04/02/26	-	ARP, ICMP echo and HTTP GET all work reliably
; 04/02/21	-	added ROM filesystem
; 04/02/18	-	first reliable result
; 04/02/14	-	added NMI latch code
; 04/02/06	-	added NIC command waiting
; 03/12/20	-	added VIC screen handling
; 03/12/14	-	added VIC interrupt handling
; 03/11/24	-	original version from the EhBASIC version

; some commented out lines have [NTSC] in their comment fields, these
; the alternate values for NTSC machines but are untested as mine are
; all PAL

; *********************************************************************
; RAM use - page 0

		*=	$0000

tp_l		.ds	1			; temp pointer low byte
tp_h		.ds	1			; temp pointer high byte

te_l		.ds	1			; temp end pointer low byte
te_h		.ds	1			; temp end pointer high byte

chk_adr_l	.ds	1			; checksum address low byte
chk_adr_h	.ds	1			; checksum address high byte

cur_l		.ds	1			; screen cursor low byte
cur_h		.ds	1			; screen cursor high byte

com_vec_l	.ds	1			; indirection vector low byte
com_vec_h	.ds	1			; indirection vector high byte

com_str_l	.ds	1			; command string pointer low byte
com_str_h	.ds	1			; command string pointer high byte

prng		.ds	1			; prng number seed

; *********************************************************************
; variables for file system

; zero page

filesys_l	.ds	1			; file system pointer low byte
filesys_h	.ds	1			; file system pointer high byte

lensys_l	.ds	1			; file system length low byte
lensys_h	.ds	1			; file system length high byte

file_l	.ds	1			; current file pointer low byte
file_h	.ds	1			; current file pointer high byte

length_l	.ds	1			; current file pointer length low byte
length_h	.ds	1			; current file pointer length high byte

fileflags	.ds	1			; current file's flag byte

findname_l	.ds	1			; filename to find pointer low byte
findname_h	.ds	1			; filename to find pointer high byte

; these locations aren't used by the program, they're just to define
; the file header structure so keep them at the end of the page zero
; definitions

f_start	.ds	1			; flags
f_length	.ds	2			; file payload length
f_name	.ds	1			; filename and null terminator byte
;		.ds	length		; here to here+length-1 is payload


; *********************************************************************
; constants for file system

none		=	$00			; no flags, no parent

; flags:

deleted	=	$80			; file is deleted but not removed
directory	=	$40			; file is really a directory
writeable	=	$01			; file is writable

; *********************************************************************
; RAM use - unexpanded memory only so far

		*=	$0200

mac_ours	.ds	6			; my mac address
mac_ours_00	=	mac_ours		; my mac address byte 00
mac_ours_01	=	mac_ours+1		; my mac address byte 01
mac_ours_02	=	mac_ours+2		; my mac address byte 02
mac_ours_03	=	mac_ours+3		; my mac address byte 03
mac_ours_04	=	mac_ours+4		; my mac address byte 04
mac_ours_05	=	mac_ours+5		; my mac address byte 05
ip_ours	.ds	4			; my IP address
seq_ours	.ds	4			; our TCP sequence number
ip_recd	.ds	4			; senders IP address
seq_recd	.ds	4			; received TCP sequence number

valid_pkt	.ds	1			; no valid packet received
IB		.ds	1			; interrupt flag byte

rx_len_l	.ds	1			; packet length low byte
rx_len_h	.ds	1			; packet length high byte

tc_l		.ds	1			; temp count low byte
tc_h		.ds	1			; temp count high byte

tl_l		.ds	1			; total length low byte
tl_h		.ds	1			; total length high byte

tx_len_l	.ds	1			; transmil length low byte
tx_len_h	.ds	1			; transmit length high byte

chksum_l	.ds	1			; checksum low byte
chksum_h	.ds	1			; checksum high byte

chk_len_l	.ds	1			; checksum length low byte
chk_len_h	.ds	1			; checksum length high byte

src_port_l	.ds	1			; source port low byte
src_port_h	.ds	1			; source port high byte

exp_port_l	.ds	1			; expected port low byte
exp_port_h	.ds	1			; expected port high byte

tcp_datap_l	.ds	1			; TCP data pointer low byte
tcp_datap_h	.ds	1			; TCP data pointer high byte

tcp_d_len_l	.ds	1			; TCP data length low byte
tcp_d_len_h	.ds	1			; TCP data length high byte

tcp_toc	.ds	1			; TCP time out count
tcp_flag	.ds	1			; TCP flags
tcp_state	.ds	1			; TCP state

tcp_ds_l	.ds	1			; TCP Tx data size low byte
tcp_ds_h	.ds	1			; TCP Tx data size high byte

aa_l		.ds	1			; data offset low byte
aa_h		.ds	1			; data offset high byte

cdtb_l	.ds	1			; countdown timer b low byte
cdtb_h	.ds	1			; countdown timer b high byte

cdtf_l	.ds	1			; countdown timer f low byte
cdtf_h	.ds	1			; countdown timer f high byte

b2b_low	.ds	1			; binary to byte temp store
logo_pos	.ds	1			; moving logo position


; *********************************************************************
; rest of the available RAM is $1000 to $1DFF

		*=	$1000

pk_buff	.ds	1514			; packet buffer
arp_buff	=	pk_buff+$0E		; ARP packet buffer
ip_buff	=	pk_buff+$0E		; IP packet buffer
tcp_buff	=	ip_buff+$14		; TCP packet buffer
icmp_buff	=	ip_buff+$14		; ICMP packet buffer

arp_ofs	=	arp_buff-pk_buff	; offset to ARP packet start
ip_ofs	=	ip_buff-pk_buff	; offset to IP packet start
tcp_ofs	=	tcp_buff-pk_buff	; offset to TCP packet start
icmp_ofs	=	icmp_buff-pk_buff	; offset to ICMP packet start

; *********************************************************************
; Ethernet header definitions

ET_dst_adr	=	$00			; destination MAC
ET_src_adr	=	$06			; source MAC
ET_type_h	=	$0C			; Type
ET_type_l	=	$0D			; Type

; *********************************************************************
; ARP header definitions

ARP_hrdsp_h	=	$00			; hardware address space identifier
ARP_hrdsp_l	=	$01			; hardware address space identifier
ARP_prtsp_h	=	$02			; protocol address space identifier
ARP_prtsp_l	=	$03			; protocol address space identifier
ARP_hrd_len	=	$04			; length of hardware address
ARP_prt_len	=	$05			; length of protocol address
ARP_code_h	=	$06			; operation code high byte
ARP_code_l	=	$07			; operation code low byte
ARP_src_mac	=	$08			; packet source MAC address
ARP_src_ip	=	$0E			; source ip address
ARP_dst_mac	=	$12			; packet target MAC address
ARP_dst_ip	=	$18			; destination ip address

; *********************************************************************
; IP header definitions

IP_version	=	$00			; version & header length
IP_service	=	$01			; type of service
IP_length_h	=	$02			; total length - high byte
IP_length_l	=	$03			; total length - low byte
IP_ident_h	=	$04			; identification - high byte
IP_ident_l	=	$05			; identification - low byte
IP_flags_h	=	$06			; flags - high byte
IP_flags_l	=	$07			; flags - low byte
IP_t2l	=	$08			; time to live
IP_protocol	=	$09			; protocol
IP_cksum_h	=	$0A			; header checksum - high byte
IP_cksum_l	=	$0B			; header checksum - low byte
IP_src_add	=	$0C			; source IP address
IP_dst_add	=	$10			; destination IP address

; IP protocol types

IP_ICMP	=	$01			; is ICMP protocol
IP_TCP	=	$06			; is TCP protocol
IP_UDP	=	$11			; is UDP protocol

; *********************************************************************
; TCP header definitions

TCP_src_p_h	=	$00			; source Port - high byte
TCP_src_p_l	=	$01			; source Port - low byte
TCP_dst_p_h	=	$02			; destination Port - high byte
TCP_dst_p_l	=	$03			; destination Port - low byte
TCP_seq_n	=	$04			; sequence Number
TCP_ack_n	=	$08			; Acknowledgment Number
TCP_d_oset	=	$0C			; data Offset
TCP_flags	=	$0D			; flags
TCP_wind_h	=	$0E			; window - high byte
TCP_wind_l	=	$0F			; window - low byte
TCP_cksum_h	=	$10			; Checksum - high byte
TCP_cksum_l	=	$11			; Checksum - low byte
TCP_urg_p_h	=	$12			; urgent Pointer - high byte
TCP_urg_p_l	=	$13			; urgent Pointer - low byte

; *********************************************************************
; ICMP header definitions

ICMP_type	=	$00			; Type
ICMP_code	=	$01			; Code
ICMP_csum_h	=	$02			; Checksum - high byte
ICMP_csum_l	=	$03			; Checksum - low byte
ICMP_id_h	=	$04			; Identifier - high byte
ICMP_id_l	=	$05			; Identifier - low byte
ICMP_seq_h	=	$06			; Sequence Number - high byte
ICMP_seq_l	=	$07			; Sequence Number - low byte

; *********************************************************************
; now the hardware bits

; ISA interface

NMI_res	=	$9800			; NMI latch reset		I/O area 2
ISA		=	$9C00			; ISA interface		I/O area 3


; 3Com 3C509B network interface card

id_port	=	ISA+$110		; id state machine port

; every page

com_reg_l	=	ISA+$30E		; command/status register low byte
com_reg_h	=	ISA+$30F		; command/status register high byte

; page 0

config_c_l	=	ISA+$304		; configuration control low byte
config_c_h	=	ISA+$305		; configuration control high byte
r_config_l	=	ISA+$308		; resource configuration low byte
r_config_h	=	ISA+$309		; resource configuration high byte
epr_ctr_l	=	ISA+$30A		; EEPROM command register low byte
epr_ctr_h	=	ISA+$30B		; EEPROM command register high byte
epr_dat_l	=	ISA+$30C		; EEPROM data register low byte
epr_dat_h	=	ISA+$30D		; EEPROM data register high byte

; page 1

data_reg	=	ISA+$300		; Tx/Rx data register
rx_stat_l	=	ISA+$308		; Rx status low byte
rx_stat_h	=	ISA+$309		; Rx status high byte
tx_stat_h	=	ISA+$30B		; Tx status high byte

; page 2

MAC_adr_00	=	ISA+$300		; MAC address 0 register
MAC_adr_01	=	ISA+$301		; MAC address 1 register
MAC_adr_02	=	ISA+$302		; MAC address 2 register
MAC_adr_03	=	ISA+$303		; MAC address 3 register
MAC_adr_04	=	ISA+$304		; MAC address 4 register
MAC_adr_05	=	ISA+$305		; MAC address 5 register

; page 4

FIFO_dag_l	=	ISA+$304		; FIFO diagnostic low byte
FIFO_dag_h	=	ISA+$305		; FIFO diagnostic high byte
media_t_l	=	ISA+$30A		; media type & status low byte


; *********************************************************************
; ROM to be used is $A000 and possibly $2000, $4000 and $6000 if needed

		*=	$A000

						; first up is the auto-boot header
	.dw	LAB_start			; entry point for code (power up)
	.dw	LAB_restart			; warm start entry point (restore key)

	.db	"A0",$C3,$C2,$CD		; 'A0CBM' boot string

LAB_restart
LAB_start
	JSR	vicinit			; initialise VIC 20 hardware

	JSR	setscr			; initialise the screen

	LDA	#$A9				; 169
	STA	ip_ours			; save my IP address first byte
	LDA	#$FE				; 254
	STA	ip_ours+$01			; save my IP address second byte
	LDX	#$00				; clear X
	STX	IB				; clear interrupt flag byte
	STX	valid_pkt			; clear valid packet
	STX	cdtb_l			; clear countdown timer b low byte
	STX	cdtb_h			; clear countdown timer b high byte
	STX	cdtf_l			; clear countdown timer f low byte
	INX					; = $01
	STX	cdtf_h			; set countdown timer f high byte
	STX	tcp_state			; set TCP state to listen

	LDA	NMI_res			; reset NMI latch
	JSR	LAB_10500			; initialise NIC

	JSR	print_ip			; print out IP address
	JSR	print_mac			; print out MAC address
	JSR	print_tcp			; print TCP state


; *********************************************************************
; main program loop

LAB_200
	LDA	IB				; get interrupt flag byte
	BEQ	LAB_990			; branch if no interrupt flagged

LAB_210
	AND	#$02				; mask adapter fail interrupt
	BEQ	LAB_220			; branch if not adapter fail interrupt

	JSR	LAB_20300			; else do adapter fail interrupt
LAB_220
	LDA	IB				; get interrupt flag byte
	AND	#$04				; mask Tx interrupt
	BEQ	LAB_225			; branch if no Tx interrupt

	JSR	LAB_20200			; else do tx interrupt
LAB_225
	LDA	IB				; get interrupt flag byte
	AND	#$10				; mask Rx interrupt
	BEQ	LAB_300			; branch if no Rx interrupt

	JSR	LAB_20100			; else do rx interrupt
LAB_300
	LDA	valid_pkt			; get valid packet flag
	BEQ	LAB_985			; skip packet check if no valid packet

						; now decipher packet info
	LDA	pk_buff+ET_type_h		; get type byte
	CMP	#$08				; compare with $08
	BNE	LAB_980			; branch if not recognised				##

	LDA	pk_buff+ET_type_l		; get type byte
	BNE	LAB_315			; branch if not IP

	JSR	LAB_21500			; handle known IP packet type
	JMP	LAB_980			; go clear packet & loop

LAB_315
	CMP	#$06				; compare with $06
	BNE	LAB_980			; branch if not recognised				##

	JSR	LAB_21000			; else handle ARP packet

LAB_980
	LDA	#$00				; clear A
	STA	valid_pkt			; clear valid packet

LAB_985
	LDA	IB				; get interrupt flag byte
	AND	#$FE				; mask only needed bits
	BNE	LAB_210			; loop while still interrupts to service

	LDX	#$FF				; bit 0 interrupt latch					##
	LDA	#$68				; acknowledge interrupt
	STX	com_reg_l			;
	STA	com_reg_h			; now ack the interrupt(s)

						; do background tasks
LAB_990
	JSR	print_tcp			; print TCP state
	JSR	periodic			; go check periodic timeout(s)
	INC	prng				; increment PRNG seed byte
	JMP	LAB_200			; loop forever


; *********************************************************************
; initialize the 3C509B

; jump through 3Com hoops - don't ask

LAB_10500
	LDY	#$00				; clear Y
	STY	id_port			; initialise id sequence state machine
	STY	id_port			; initialise id sequence state machine

						; send id sequence to NIC id port
	DEY					; set for 255 bytes
	TYA					; set initial byte
LAB_10520
	STA	id_port			; send id sequence state machine byte
	ASL					; shift byte
	BCC	LAB_10535			; branch if no carry

	EOR	#$CF				; else EOR with $CF
LAB_10535
	DEY					; increment count
	BNE	LAB_10520			; loop if not all done
	
						; no contention check, only 1 card!
	LDA	#$D1				; set tag to 1
	STA	id_port			; send id port byte
	LDA	#$F0				; set I/O base to $0300
	STA	id_port			; send id port byte

	LDA	#$08				;
	STY	com_reg_l			; $00 page 0
	STA	com_reg_h			; $08 set window command

						; read our MAC address from the card
	LDX	#$00				; clear index
LAB_10560
	TXA					; copy index
	LSR					; /2 (EEPROM is read as word wide device)
	ORA	#$80				; OR with EEPROM read command
	STA	epr_ctr_l			; set EEPROM read address low byte
	STY	epr_ctr_h			; set EEPROM read address high byte

						; wait for at least 162 uS
	LDY	#$24				; [2] cycles plus .. [PAL]
;##	LDY	#$21				; [2] cycles plus .. [NTSC]
LAB_10565
	DEY					; [2] cycles plus ..
	BNE	LAB_10565			; [3/2] cycles = 5 cycles times Y plus 1 cycle
						; = 181 cycles = 163uS on a PAL machine
						; = 166 cycles = 162uS on an NTSC machine

	LDY	epr_dat_l			; get MAC address low byte
	LDA	epr_dat_h			; get MAC address high byte
	STA	mac_ours,X			; save MAC address high byte
	INX					; increment index
	TYA					; copy MAC address low byte
	STA	mac_ours,X			; save MAC address low byte
	LDY	#$00				; clear Y
	INX					; increment index
	CPX	#$06				; are we done yet
	BNE	LAB_10560			; loop if not

						; enable the adaptor
	LDA	#$01				;
 	STA	config_c_l			; $01
 	STY	config_c_h			; $00 set configuration control

						; reset the Tx and Rx and suspend status
	LDA	#$58				;
	STY	com_reg_l			; $00
	STA	com_reg_h			; $58 Tx reset
	JSR	wait_cmd			; wait for Tx reset to complete

	LDA	#$28				;
	STY	com_reg_l			; $00
	STA	com_reg_h			; $28 Rx reset 
	JSR	wait_cmd			; wait for Rx reset to complete

	LDA	#$78				;
	STY	com_reg_l			; $00
	STA	com_reg_h			; $78 mask all status bits

						; set IRQ3
	LDA	#$3F				;
 	STY	r_config_l			; $00
 	STA	r_config_h			; $3F resource configuration
	
						; set our MAC address & make MAC low word
						; our IP address low word
	LDA	#$02				;
	STA	com_reg_l			;
	LDA	#$08				;
	STA	com_reg_h			; set page 2

	LDA	mac_ours_00			; get MAC address byte 0
	STA	MAC_adr_00			; save to card
	LDA	mac_ours_01			; get MAC address byte 1
	STA	MAC_adr_01			; save to card
	LDA	mac_ours_02			; get MAC address byte 2
	STA	MAC_adr_02			; save to card
	LDA	mac_ours_03			; get MAC address byte 3
	STA	MAC_adr_03			; save to card
	LDA	mac_ours_04			; get MAC address byte 4
	STA	MAC_adr_04			; save to card
	STA	ip_ours+$02			; save IP byte
	LDA	mac_ours_05			; get MAC address byte 5
	STA	MAC_adr_05			; save to card
	STA	ip_ours+$03			; save IP byte

						; enable link beat & jabber for TP
	LDA	#$04				;
	STA	com_reg_l			;
	LDA	#$08				;
	STA	com_reg_h			; set page 4

	LDA	#$C0				;
	STA	media_t_l			; set media type & status (only low byte needed)

	LDA	#$B0				;
	STY	com_reg_l			; $00
	STA	com_reg_h			; $B0 stats disable

	LDA	#$06				;
	STA	com_reg_l			;
	LDA	#$08				;
	STA	com_reg_h			; set page 6

						; clear stats by reading them
	LDX	#$00				; clear index
LAB_10655
	LDA	data_reg,X			; read stat byte
	INX					; increment index
	CPX	#$0E				; are we done yet
	BNE	LAB_10655			; loop if not

	LDA	#$A8				;
	STY	com_reg_l			; $00
	STA	com_reg_h			; $A8 stats enable

						; set to receive host address and broadcasts
	LDA	#$01				;
	STA	com_reg_l			;
	LDA	#$08				;
	STA	com_reg_h			; set page 1

	LDA	#$05				;
	STA	com_reg_l			;
	LDA	#$80				;
	STA	com_reg_h			; set Rx filter

						; enable Rx and Tx
	LDA	#$20				;
	STY	com_reg_l			; $00
	STA	com_reg_h			; $20 Rx enable

	LDA	#$48				;
	STY	com_reg_l			; $00
	STA	com_reg_h			; $48 Tx enable

						; enable those status bits that we want readable
	LDA	#$16				; bit 4 Rx complete
						; bit 2 Tx complete
						; bit 1 adapter failure
						; bit 0 interrupt latch
	STA	com_reg_l			;
	LDA	#$78				;
	STA	com_reg_h			; set read zero mask

						; ack any pending interrupts
	LDA	#$69				; bit 6 interrupt requested
						; bit 5 rx early
						; bit 3 tx available
						; bit 0 interrupt latch
	STA	com_reg_l			;
	LDA	#$68				;
	STA	com_reg_h			; ack interrupts

						; enable those interrupts that we want,
	LDA	#$16				; bit 4 Rx complete
						; bit 2 Tx complete
						; bit 1 adapter failure
						; bit 0 interrupt latch
	STA	com_reg_l			;
	LDA	#$70				;
	STA	com_reg_h			; set interrupt mask

	RTS

; *********************************************************************
; when this was all done in EhBASIC it was easy to blindly issue
; commands to the NIC and assume they would be finished before the
; next command was issued. however in assembly it is easy to issue a
; new command before some commands have completed so to fix this the
; status register is polled repeatedly until the command in progress
; bit clears. note - this is only needed after some commands.
; *********************************************************************
; the above sounds like agood plan but you must disable interrupts
; while polling and if an interrupt condition is met while interrupts
; are disabled you'll miss it unless you poll the interrupt bits as
; well. so for now just wait longer than the longest command takes.
; *********************************************************************
; wait for an NIC command to complete

wait_cmd
	CLC					; clear carry for add
	LDA	#$80				; count $80
wait_loop
	ADC	#$01
	BCC	wait_loop

	RTS


; *********************************************************************
; does the timeout counting down, called on TIMER 1 timeout (1mS)
; decrement the TCP timeout counter if <> 0

LAB_12000
	LDA	tcp_toc			; test TCP timeout count
	BEQ	LAB_12005			; skip decrement if zero

	DEC	tcp_toc			; else decrement it
LAB_12005
	LDA	cdtb_l			; get countdown timer b low byte
	BEQ	LAB_12010			; skip decrement if zero

	DEC	cdtb_l			; else decrement it
LAB_12010
	LDA	cdtf_l			; get countdown timer f low byte
	BEQ	LAB_12015			; skip decrement if zero

	DEC	cdtf_l			; else decrement it
LAB_12015
	RTS


; *********************************************************************
; handle Rx interrupt flagged, get the ethernet packet from the NIC
; exit with length in rx_len, payload in pk_buff(). valid_pkt flags the
; packet status

LAB_20100
	LDX	#$00				;
	STX	valid_pkt			; clear packet valid flag

	LDA	#$EF				; xxx0 xxxx, set for Rx interrupt bit
	AND	IB				; clear IB bit(s)
	STA	IB				; save IB

	LDX	#$01				;
	LDA	#$08				;
	STX	com_reg_l			; $01
	STA	com_reg_h			; $08 set page 1

	LDA	rx_stat_l			; get Rx status low byte
	STA	tc_l				; save length to temp count low byte

	LDA	rx_stat_h			; get Rx status high byte
	CMP	#$40				; test error bit
	BCS	LAB_20145			; discard packet & exit if Rx error

	AND	#$07				; else mask length bits
	STA	tc_h				; save length to temp count high byte
	CMP	#$05				; compare with max length high byte
	BCC	LAB_20135			; branch if ok (less)

	BNE	LAB_20145			; discard packet & exit if too big

	LDA	tc_l				; else get length low byte
	CMP	#$DD				; compare with max+1
	BCS	LAB_20145			; discard packet & exit if too big

; get packet from NIC

LAB_20135
	LDA	#>pk_buff			; get IP buffer pointer high byte
	STA	tp_h				; save to temp pointer high byte
	LDA	#<pk_buff			; get IP buffer pointer low byte
	STA	tp_l				; save to temp pointer low byte
	LDY	#$00				; clear index
LAB_20136
	LDA	data_reg			; get byte from NIC
	STA	(tp_l),Y			; save to IP buffer
	INY					; increment index (pointer low byte)
	BNE	LAB_20137			; branch if no overflow

	INC	tp_h				; else increment pointer high byte
LAB_20137
	DEC	tc_l				; decrement count low byte
	BNE	LAB_20136			; loop if not all done

	DEC	tc_h				; else decrement count high byte
	BPL	LAB_20136			; loop if not all done

	DEC	valid_pkt			; flag this packet good
LAB_20145
	LDX	#$00				;
	LDA	#$40				;
	STX	com_reg_l			; $00
	STA	com_reg_h			; $40 discard top packet
	JSR	wait_cmd			; wait for discard to complete

	RTS


; *********************************************************************
; handle Tx interrupt flagged

LAB_20200
	LDA	#$FB				; xxxx x0xx, set for Tx interrupt bit
	AND	IB				; clear IB bit(s)
	STA	IB				; save IB

	LDX	#$01				;
	LDA	#$08				;
	STX	com_reg_l			; $01
	STA	com_reg_h			; $08 set page 1

	LDA	tx_stat_h			; read Tx status (high byte only)
	AND	#$3C				; mask all Tx errors
	BEQ	LAB_20220			; branch if no errors

	DEX					; clear X
	AND	#$30				; mask jabber or underrun
	BEQ	LAB_20215			; branch if no jabber or underrun

						; do a Tx reset if jabber or underrun was indicated
	LDA	#$58				;
 	STX	com_reg_l			; $00
	STA	com_reg_h			; $58 Tx reset
	JSR	wait_cmd			; wait for Tx reset to complete

LAB_20215					; do a Tx enable if any errors were indicated
	LDA	#$48				;
 	STX	com_reg_l			; $00
	STA	com_reg_h			; $48 Tx enable

LAB_20220
 	STX	tx_stat_h			; ack Tx interrupts by writing to status (any value)
	RTS


; *********************************************************************
; handle adapter fail interrupt flagged

LAB_20300
	LDA	#$FD				; xxxx xx0x, set for adapter fail interrupt flag
	AND	IB				; clear IB bit(s)
	STA	IB				; save IB

	LDX	#$04				;
	LDA	#$08				;
	STX	com_reg_l			; $04
	STA	com_reg_h			; $08 page 4

	LDX	FIFO_dag_h			; get FIFO diagnostic high byte (fail reason)

	LDY	#$01				;
	STY	com_reg_l			; $01
	STA	com_reg_h			; $08 page 1

	TXA					; get fail reason back
	AND	#$20				; test Rx underflow
	BEQ	LAB_20330			; branch if no Rx underflow

						; was Rx underflow so reset & restart receiver
	LDA	#$EF				; xxx0 xxxx, set for Rx interrupt bit
	AND	IB				; clear IB bit(s)
	STA	IB				; save IB

	DEY					; Y is now $00
	LDA	#$28				;
	STY	com_reg_l			; $00
	STA	com_reg_h			; $28 Rx reset
	JSR	wait_cmd			; wait for Rx reset to complete

	LDA	#$20				;
	STY	com_reg_l			; $00
	STA	com_reg_h			; $20 Rx enable

LAB_20330
	TXA					; get fail reason back
	AND	#$04				; test Tx overflow
	BEQ	LAB_20350			; branch if no Tx overflow

						; was Tx overflow so reset & restart transmitter
	LDA	#$58				;
	STY	com_reg_l			; $00
	STA	com_reg_h			; $58 Tx reset
	JSR	wait_cmd			; wait for Tx reset to complete

	LDA	#$48
	STY	com_reg_l			; $00
	STA	com_reg_h			; $48 Tx enable
LAB_20350
	RTS

; *********************************************************************
; construct a reply to an ARP request aimed at us
; see RFC826 for a description of the ARP packet

LAB_21000
	LDA	arp_buff+ARP_hrdsp_l	; hardware type - always $01
	CMP	#$01				; check type
	BNE	LAB_20350			; exit if incorrect

	LDA	arp_buff+ARP_prtsp_h	; hardware protocol high byte
	CMP	#$08				; check protocol
	BNE	LAB_20350			; exit if incorrect

	LDA	arp_buff+ARP_prtsp_l	; hardware protocol low byte
	BNE	LAB_20350			; exit if incorrect

	LDA	arp_buff+ARP_hrd_len	; hardware (MAC) address length
	CMP	#$06				; check length
	BNE	LAB_20350			; exit if incorrect

	LDA	arp_buff+ARP_prt_len	; protocol (IP) address length
	CMP	#$04				; check length
	BNE	LAB_20350			; exit if incorrect

	LDA	arp_buff+ARP_code_l	; type of packet
	CMP	#$01				; check type
	BNE	LAB_20350			; exit if incorrect

	LDX	#arp_ofs+ARP_dst_ip	; index to protocol destination address
	JSR	match_ip			; compare with our IP address
	BNE	LAB_20350			; exit if not for us

; we have recognised the packet as correct and the target as us
; set up the buffer for transmit

						; do MAC addresses
	LDX	#$05				; set index/count
LAB_21050
	LDA	arp_buff+ARP_src_mac,X	; copy source MAC address byte
	STA	pk_buff+ET_dst_adr,X	; save destiantion MAC address byte
	STA	arp_buff+ARP_dst_mac,X	; save destiantion MAC address byte
	LDA	mac_ours,X			; get our MAC address byte
	STA	pk_buff+ET_src_adr,X	; save source MAC address byte
	STA	arp_buff+ARP_src_mac,X	; save source MAC address byte
	DEX
	BPL	LAB_21050			; loop if not all done

						; do IP addresses
	LDX	#$03				; set index/count
LAB_21055
	LDA	arp_buff+ARP_src_ip,X	; get source IP address byte
	STA	arp_buff+ARP_dst_ip,X	; save destination IP address byte
	LDA	ip_ours,X			; get our IP address byte
	STA	arp_buff+ARP_src_ip,X	; save source IP address byte
	DEX					; increment index
	BPL	LAB_21055			; loop if not all done

						; make it an ARP reply
	LDA	#$02				; set op type to reply
	STA	arp_buff+ARP_code_l	; save op type

	LDA	#$2A				; ARP reply length low byte
	STA	tx_len_l			; set Tx length low byte
	LDA	#$00				; ARP reply length high byte
	STA	tx_len_h			; set Tx length high byte

						; packet set up, now Tx it

; *********************************************************************
; Tx ethernet packet, tx_len is length

LAB_21200
	LDA	tx_len_l			; get Tx length low byte
	STA	tc_l				; save to temp count
	STA	data_reg			; save to NIC
	LDA	tx_len_h			; get Tx length high byte
	STA	tc_h				; save to temp count
	ORA	#$80				; generates an interrupt when sent
	STA	data_reg			; save to NIC
	LDY	#$00				; clear index
	STY	data_reg			; save to NIC
	STY	data_reg			; save to NIC

						; header written, now write data
	LDA	#>pk_buff			; get buffer pointer high byte
	STA	tp_h				; save to temp pointer high byte
	LDA	#<pk_buff			; get IP buffer pointer low byte
	STA	tp_l				; save to temp pointer low byte
LAB_22105
	LDA	(tp_l),Y			; get byte from IP buffer
	STA	data_reg			; save to NIC
	INY					; increment index (pointer low byte)
	BNE	LAB_22110			; branch if no overflow

	INC	tp_h				; else increment pointer high byte
LAB_22110
	DEC	tc_l				; decrement count low byte
	BNE	LAB_22105			; loop if not all done

	DEC	tc_h				; else decrement count high byte
	BPL	LAB_22105			; loop if not all done

						; now calculate longword remainder
	LDA	tx_len_l			; get count low byte
	AND	#$03				; do MOD 4
	BEQ	LAB_21220			; exit if no remainder

						; else pad to next longword
	EOR	#$03				; do 3-MOD
	TAX					; copy to index
	LDA	#$00				; clear A
LAB_21215
	STA	data_reg			; save to NIC
	DEX					; decrement count
	BPL	LAB_21215			; loop if not all done

LAB_21220
	RTS


; *********************************************************************
; match IP address, X = buffer index, returns Zb=1 if ours

match_ip
	LDY	#$FC				; set count to -4
match_loop
	LDA	pk_buff,X			; get protocol address byte
	CMP	ip_ours-$FC,Y		; compare with our IP address byte
	BNE	match_exit			; exit if not for us

	INX					; increment index
	INY					; increment count
	BNE	match_loop			; loop if more to do

match_exit
	RTS


; *********************************************************************
; IP protocol types and vectors

; more protocol types can be accomodated by adding the type byte to the
; IPprot table and the low and high vetor bytes for the code to the
; IPpvec tables. as an example the commented out entries for UDP are
; included. protocol code should exit via an RTS

IPprot
	.byte	IP_ICMP			; ICMP type
	.byte	IP_TCP			; TCP type
;	.byte	IP_UDP			; UDP type
IPpend

IPpvec_l
	.byte	<LAB_21600			; ICMP code vector low byte
	.byte	<LAB_22000			; TCP code vector low byte
;	.byte	<LAB_UDP			; UDP code vector low byte

IPpvec_h
	.byte	>LAB_21600			; ICMP code vector high byte
	.byte	>LAB_22000			; TCP code vector high byte
;	.byte	>LAB_UDP			; UDP code vector high byte


; *********************************************************************
; handle IP packet
; check the type, fragmentation and checksum of an incomming IP packet.
; ignore tyes we don't recognise or fragments or broken packets. pass
; the packets on to the protocol handlers

LAB_21500
	LDA	ip_buff+IP_version	; get IP header length & type
	CMP	#$45				; compare with expected ..
						; type 4
						; length 5 longwords
	BNE	LAB_21565			; don't recognise anything else

	LDX	#ip_ofs+IP_dst_add	; index to IP address
	JSR	match_ip			; compare with our IP address
	BNE	LAB_21565			; exit if not ours

	LDA	ip_buff+IP_flags_h	; get flags+offset high byte
	AND	#$3F				; mask wanted bits
	BNE	LAB_21565			; we don't handle fragments (yet)

	LDA	ip_buff+IP_flags_l	; get offset low byte
	BNE	LAB_21565			; we don't handle fragments (yet)

	STA	chksum_l			; clear checksum low byte
	STA	chksum_h			; clear checksum high byte

	LDX	#ip_ofs			; offset to start of IP payload
	STX	chk_adr_l			; set checksum address low byte
	STA	chk_adr_h			; clear checksum address high byte

	LDX	#$14				; IP header length
	STX	chk_len_l			; set checksum length low byte
	STA	chk_len_h			; clear checksum length high byte

	JSR	LAB_21950			; do checksum
	LDA	chksum_l			; get checksum low byte
	AND	chksum_h			; AND checksum high byte
	TAX					; copy it
	INX					; test checksum
	BNE	LAB_21565			; ignore broken checksum

	LDA	ip_buff+IP_protocol	; get IP protocol
	LDX	IPpend-IPprot-1		; length of table -1
LAB_IPprotloop
	CMP	IPprot,X			; compare with protocol type
	BEQ	LAB_doIPprot		; if match go do protocol

	DEX					; decrement index
	BPL	LAB_IPprotloop		; loop if not all checked

LAB_21565
	RTS

; protocol matched so get the vector and do it

LAB_doIPprot
	LDA	IPpvec_l,X			; get vector low byte
	STA	com_vec_l			; save vector low byte
	LDA	IPpvec_h,X			; get vector high byte
	STA	com_vec_h			; save vector high byte
	JMP	(com_vec_l)			; do vector


; *********************************************************************
; handle ICMP - we only recognise the echo request packet
; this is the bit that goes PING!

LAB_21600
	JSR	LAB_21605			; do ICMP checksum
	LDA	chksum_l			; get checksum low byte
	AND	chksum_h			; AND checksum high byte
	TAX					; copy it
	INX					; test checksum
	BNE	LAB_21565			; ignore broken checksum

	LDA	icmp_buff+ICMP_type	; get ICMP type
	EOR	#$08				; compare with echo request
	BNE	LAB_21565			; ignore any other type

	STA	icmp_buff+ICMP_type	; set ICMP type (reply)
	STA	icmp_buff+ICMP_code	; set ICMP code

	STA	icmp_buff+ICMP_csum_h	; clear ICMP checksum high byte
	STA	icmp_buff+ICMP_csum_l	; clear ICMP checksum low byte

	JSR	LAB_21605			; calculate new ICMP checksum

	LDA	chksum_h			; get checksum high byte
	EOR	#$FF				; ones complement checksum high byte
	STA	icmp_buff+ICMP_csum_h	; save ICMP checksum high byte
	LDA	chksum_l			; get checksum low byte
	EOR	#$FF				; ones complement checksum low byte
	STA	icmp_buff+ICMP_csum_l	; save ICMP checksum low byte

	JSR	LAB_21800			; setup the IP header

						; now  set the length and transmit the packet
	CLC					; clear carry for add
	LDA	ip_buff+IP_length_l	; get IP length low byte
	ADC	#<ip_ofs			; add ethernet header length
	STA	tx_len_l			; save Tx length low byte

	LDA	ip_buff+IP_length_h	; get data length high byte
	ADC	#>ip_ofs			; add carry
	STA	tx_len_h			; save Tx length high byte

	JMP	LAB_21200			; transmit packet & return


; *********************************************************************
; do ICMP IP echo request packet checksum

LAB_21605
	LDA	#$00				; clear A
	STA	chksum_l			; clear checksum low byte
	STA	chksum_h			; clear checksum high byte
	STA	chk_adr_h			; set checksum address high byte

	LDA	#$22				; start of ICMP reply
	STA	chk_adr_l			; set checksum address low byte

	SEC					; set carry for subtract
	LDA	ip_buff+IP_length_l	; get data length low byte
	SBC	#$14				; subtract header length
	STA	chk_len_l			; save checksum length low byte

	LDA	ip_buff+IP_length_h	; get data length high byte
	SBC	#$00				; subtract carry
	STA	chk_len_h			; save checksum length high byte

	JMP	LAB_21950			; do checksum & return


; *********************************************************************
; swap the sources and destinations (MAC and IP) and compute the new IP
; header checksum

LAB_21800
	LDX	#$05				; set index/count
LAB_21805
	LDA	pk_buff+ET_src_adr,X	; copy source MAC address byte
	STA	pk_buff+ET_dst_adr,X	; save destiantion MAC address byte
	LDA	mac_ours,X			; get our MAC address byte
	STA	pk_buff+ET_src_adr,X	; save source MAC address byte
	DEX
	BPL	LAB_21805			; loop if not all done

						; do IP addresses
	LDX	#$03				; set index/count
LAB_21820
	LDA	ip_buff+IP_src_add,X	; get source IP address byte
	STA	ip_buff+IP_dst_add,X	; save destination IP address byte
	LDA	ip_ours,X			; get our IP address byte
	STA	ip_buff+IP_src_add,X	; save source IP address byte
	DEX					; increment index
	BPL	LAB_21820			; loop if not all done

	LDA	#$80				; set time to live
	STA	ip_buff+IP_t2l		; save time to live

						; calculate the IP header checksum
	LDA	#$00				; clear A
	STA	ip_buff+IP_cksum_h	; clear packet checksum high byte
	STA	ip_buff+IP_cksum_l	; clear packet checksum low byte
	STA	chksum_l			; clear calculated checksum low byte
	STA	chksum_h			; clear calculated checksum high byte
	STA	chk_len_h			; clear checksum length high byte
	STA	chk_adr_h			; clear checksum address high byte

	LDA	#$14				; header length
	STA	chk_len_l			; save checksum length low byte

	LDA	#$0E				; IP start
	STA	chk_adr_l			; save checksum address low byte

	JSR	LAB_21950			; do checksum

	LDA	chksum_h			; get checksum high byte
	EOR	#$FF				; ones complement
	STA	ip_buff+IP_cksum_h	; save checksum high byte

	LDA	chksum_l			; get checksum low byte
	EOR	#$FF				; ones complement
	STA	ip_buff+IP_cksum_l	; save checksum low byte

	RTS


; *********************************************************************
; compute or check the 16 bit checksum on the TCP part of the packet

LAB_21900
	SEC					; set carry for subtract
	LDA	ip_buff+IP_length_l	; get IP total length low byte
	SBC	#tcp_ofs-ip_ofs		; subtract IP header length
	STA	tl_l				; save TCP total length low byte
	STA	chksum_l			; checksum is TCP length

	LDA	ip_buff+IP_length_h	; get IP total length high byte
	SBC	#$00				; subtract carry
	STA	tl_h				; save TCP total length high byte
	STA	chksum_h			; save checksum high byte

	CLC					; clear carry for add
	LDA	ip_buff+IP_protocol	; get protocol
	ADC	chksum_l			; add checksum low byte
	STA	chksum_l			; save checksum low byte
	LDA	chksum_h			; get checksum high byte
	ADC	#$00				; add carry
	STA	chksum_h			; save checksum high byte

	LDA	#$08				; set checksum length low byte
	STA	chk_len_l			; save checksum length low byte

	LDA	#ip_ofs+IP_src_add	; set checksum address low byte
	STA	chk_adr_l			; save checksum address low byte

	LDA	#$00				; clear A
	STA	chk_len_h			; set checksum length high byte
	STA	chk_adr_h			; save checksum address high byte

	JSR	LAB_21950			; do IP address checksum

	LDA	tl_l				; get TCP length low byte
	STA	chk_len_l			; save checksum count low byte
	LDA	tl_h				; get TCP length high byte
	STA	chk_len_h			; save checksum count high byte

	LDA	#<tcp_ofs			; set TCP start
	STA	chk_adr_l			; save checksum address low byte
	LDA	#>tcp_ofs			; clear A
	STA	chk_adr_h			; save checksum address high byte


; *********************************************************************
; compute the 1's complement 16 bit checksum on part of the ip() packet
; entry is with chk_len = bytes to checksum and chk_adr = index to the
; first byte

; this routine caused soooo much grief for a number of reasons, a 16
; bit byte count with odd or even exit points, a 16 bit sum where the
; bytes appear in high/low order and where the carry wraps from b15 to
; b0. also the byte count can be zero and we don't want the routine to
; take an age. I think it all works now.

LAB_21950
	CLC					; clear carry for add
	LDA	chk_adr_l			; get checksum pointer low byte
	ADC	#<pk_buff			; add buffer pointer low byte
	STA	chk_adr_l			; set checksum pointer low byte
	LDA	chk_adr_h			; get checksum pointer high byte
	ADC	#>pk_buff			; add buffer pointer high byte
	STA	chk_adr_h			; set checksum pointer high byte

	LDY	#$00				; clear index
	CPY	chk_len_l			; compare checksum length low byte
	CLC					; clear carry for (either) add
	BEQ	LAB_21972			; branch if $xx00 length

LAB_21955
	LDA	(chk_adr_l),Y		; get byte
	TAX					; copy high byte

	DEC	chk_len_l			; decrement count low byte
	BNE	LAB_21960			; branch if no rollunder

	DEC	chk_len_h			; else decrement count high byte
	BMI	LAB_21975			; exit loop if no more bytes to checksum

LAB_21960
	INY					; increment index (address low byte)
	BNE	LAB_21965			; branch if no rollover

	INC	chk_adr_h			; else increment high byte
LAB_21965
	LDA	(chk_adr_l),Y		; get byte
	ADC	chksum_l			; add checksum low byte
	STA	chksum_l			; save checksum low byte
	TXA					; get high byte
	ADC	chksum_h			; add checksum high byte
	STA	chksum_h			; save checksum high byte

	INY					; increment index (address low byte)
	BNE	LAB_21970			; branch if no rollover

	INC	chk_adr_h			; else increment high byte
LAB_21970
	DEC	chk_len_l			; decrement count low byte
	BNE	LAB_21955			; loop if no rollunder

LAB_21972
	DEC	chk_len_h			; else decrement count high byte
	BPL	LAB_21955			; loop while bytes to checksum

	BMI	LAB_21980			; just go add carry

LAB_21975
	LDA	#$00				; clear low byte
	ADC	chksum_l			; add checksum low byte
	STA	chksum_l			; save checksum low byte
	TXA					; get high byte
	ADC	chksum_h			; add checksum high byte
	STA	chksum_h			; save checksum high byte
LAB_21980
	LDA	chksum_l			; get checksum low byte
	ADC	#$00				; add any carry
	STA	chksum_l			; save checksum low byte
	LDA	chksum_h			; get checksum high byte
	ADC	#$00				; add any carry
	STA	chksum_h			; save checksum high byte
LAB_21990
	RTS


; *********************************************************************
; handle TCP packet

; tl = total length of IP packet
; tcp_datap = TCP data pointer
; tcp_d_len = TCP data length
; tcp_flag  = TCP flags for this packet
; tcp_state = TCP state
;		1 = listen
;		2 = SYN received
;		3 = established
;		4 = FIN wait 1
;		5 = FIN wait 1
; destination port MUST be $0080 (http:// only)

LAB_22000
	JSR	LAB_21900			; do TCP checksum
	LDA	chksum_l			; get checksum low byte
	AND	chksum_h			; AND checksum high byte
	TAX					; copy it
	INX					; test checksum
	BNE	LAB_22098			; ignore broken checksum

	SEC					; set carry for subtract
	LDA	ip_buff+IP_length_l	; get IP total length low byte
	STA	tl_l				; save total length low byte
	SBC	#$14				; subtract IP header length low byte
	STA	tcp_d_len_l			; save TCP data length low byte

	LDA	ip_buff+IP_length_h	; get IP total length high byte
	STA	tl_h				; save total length high byte
	SBC	#$00				; subtract carry from high byte
	STA	tcp_d_len_h			; save TCP data length high byte

	LDA	tcp_buff+TCP_src_p_h	; get source port address high byte
	STA	src_port_h			; save source port address high byte
	LDA	tcp_buff+TCP_src_p_l	; get source port address low byte
	STA	src_port_l			; save source port address low byte

	LDA	tcp_buff+TCP_dst_p_h	; get destination port high byte
	BNE	LAB_22098			; ignore ports > $FF

	LDA	tcp_buff+TCP_dst_p_l	; get destination port low byte
	CMP	#$50				; is it port 80
	BNE	LAB_22098			; ignore ports <> $50

	LDA	tcp_buff+TCP_d_oset	; TCP data offset
	AND	#$F0				; mask xxxx 0000, offset bits
	LSR					; effectively /16 ..
	LSR					; .. *4
	STA	aa_l				; copy data offset

	ADC	#<tcp_ofs			; add TCP start offset (carry clear)
	STA	tcp_datap_l			; save TCP data start
	LDA	#>tcp_ofs			; clear A
	STA	tcp_datap_h			; save TCP data start high byte

	SEC					; set carry for subtract
	LDA	tcp_d_len_l			; get TCP data length low byte
	SBC	aa_l				; subtract TCP data offset low byte
	STA	tcp_d_len_l			; save TCP data length low byte

	LDA	tcp_d_len_h			; get TCP data length high byte
	SBC	#$00				; subtract carry from high byte
	STA	tcp_d_len_h			; save TCP data length high byte

	LDA	tcp_buff+TCP_flags	; get the TCP flags
	AND	#$3F				; just TCP flags (mask reserved bits)
	STA	tcp_flag			; save TCP flags

	LDX	tcp_state			; get TCP state
	CPX	#$06				; compare with max+1
	BCS	LAB_22075			; TCP state is wrong so reset it

	LDA	LAB_2209_l,X		; get vector low byte
	STA	tp_l				; save to temp pointer low byte
	LDA	LAB_2209_h,X		; get vector high byte
	STA	tp_h				; save to temp pointer high byte
	JMP	(tp_l)			; do vector

; else TCP state is wrong so reset it

LAB_22075
	LDX	#$01				; state to listen
	STX	tcp_state			; set TCP state
LAB_22098
	RTS

						; TCP code vectors, low bytes
LAB_2209_l
	.byte	<LAB_22075			; closed (should never happen but do passive open)
	.byte	<LAB_22100			; listen
	.byte	<LAB_22200			; synrx
	.byte	<LAB_22300			; estab
	.byte	<LAB_22400			; fwait1
	.byte	<LAB_22500			; fwait2

						; TCP code vectors, high bytes
LAB_2209_h
	.byte	>LAB_22075			; closed
	.byte	>LAB_22100			; listen
	.byte	>LAB_22200			; synrx
	.byte	>LAB_22300			; estab
	.byte	>LAB_22400			; fwait1
	.byte	>LAB_22500			; fwait2


; *********************************************************************
; TCP listen state

LAB_22100
	LDA	tcp_flag			; get TCP flags
	AND	#$04				; check RST flag
	BNE	LAB_22198			; ignore RST (correct)

	LDA	tcp_flag			; get TCP flags
	AND	#$10				; check ACK flag
	BNE	LAB_22198			; ignore ACK (should send RST)

	LDA	tcp_flag			; get TCP flags
	AND	#$02				; check SYN flag
	BEQ	LAB_22198			; ignore not SYN

	LDX	#$03				; set index
LAB_22115
	LDA	tcp_buff+TCP_seq_n,X	; get rx sequence no
	STA	seq_recd,X			; save rx sequence no

	LDA	ip_buff+IP_src_add,X	; get sender's IP addr
	STA	ip_recd,X			; save sender's IP addr

	DEX					; decrement index
	BPL	LAB_22115			; loop if not all done

	JSR	LAB_get_prng		; get PRNG number
	STA	seq_ours			; save as our sequence #
	JSR	LAB_get_prng		; get PRNG number
	STA	seq_ours+$01		; save as our sequence #
	LDA	#$FF				; set to -1
	STA	seq_ours+$02		; save as our sequence #
	STA	seq_ours+$03		; save as our sequence #

						; got details, now make SYN ACK
	JSR	LAB_22815			; prepare TCP reply
	JSR	LAB_22920			; generate ACK for SYN received
	JSR	LAB_22895			; generate our own SYN
	JSR	LAB_22845			; finalise TCP reply
	JSR	LAB_21800			; generate IP header
	JSR	LAB_21200			; transmit packet

						; connected port = source port
	LDA	src_port_l			; get senders port low byte
	STA	exp_port_l			; save expected port low byte
	LDA	src_port_h			; get senders port high byte
	STA	exp_port_h			; save expected port high byte

	LDA	#$02				; TCP state to SYN received
	STA	tcp_state			; save TCP state

	LDA	#$C8				; set timeout count
	STA	tcp_toc			; save timeout count
LAB_22198
	RTS


; *********************************************************************
; TCP syn received state, we're expecting an ACK of our SYN

LAB_22200
	LDA	tcp_toc			; get TCP time out count
	BNE	LAB_22205			; branch if not timed out

	LDA	#$01				; TCP state to listen
	STA	tcp_state			; save TCP state
	RTS

LAB_22205
	LDX	#$03				; set index
LAB_22210
	LDA	ip_recd,X			; get TCP sender's IP address byte
	CMP	ip_buff+IP_src_add,X	; compare with received IP address
	BNE	LAB_22298			; exit if not expected address

	DEX					; decrement index
	BPL	LAB_22210			; loop if not all done

	LDA	src_port_l			; get sender's port low byte
	CMP	exp_port_l			; compare with expected port low byte
	BNE	LAB_22298			; exit if wrong port

	LDA	src_port_h			; get sender's port high byte
	CMP	exp_port_h			; compare with expected port high byte
	BNE	LAB_22298			; exit if wrong port

	LDX	#$03				; set index
LAB_22220
	LDA	seq_ours,X			; get sender's ack number byte
	CMP	tcp_buff+TCP_ack_n,X	; compare with received ack number
	BNE	LAB_22298			; exit if not expected ack

	DEX					; decrement index
	BPL	LAB_22220			; loop if not all done

	LDA	tcp_flag			; get TCP flags
	AND	#$02				; check SYN flag
	BNE	LAB_22298			; ignore SYN (should RST & close)

	LDA	tcp_flag			; get TCP flags
	AND	#$10				; check ACK flag
	BEQ	LAB_22298			; was not ACK

	LDA	#$03				; TCP state established
	STA	tcp_state			; save TCP state

	LDA	#$C8				; set TCP timeout count
	STA	tcp_toc			; save timeout count
LAB_22298
	RTS


; *********************************************************************
; TCP established state

LAB_22300
	LDA	tcp_toc			; get TCP time out count
	BNE	LAB_22305			; branch if not timed out

	LDA	#$01				; TCP state to listen
	STA	tcp_state			; save TCP state
	RTS

LAB_22305
	LDX	#$03				; set index
LAB_22310
	LDA	ip_recd,X			; get TCP sender's IP address byte
	CMP	ip_buff+IP_src_add,X	; compare with received IP address
	BNE	LAB_22298			; exit if not expected address

	DEX					; decrement index
	BPL	LAB_22310			; loop if not all done

	LDA	src_port_l			; get sender's port low byte
	CMP	exp_port_l			; compare with expected port low byte
	BNE	LAB_22298			; exit if wrong port

	LDA	src_port_h			; get sender's port high byte
	CMP	exp_port_h			; compare with expected port high byte
	BNE	LAB_22298			; exit if wrong port

	JSR	LAB_22800			; ck rx sequence number
	BNE	LAB_22298			; exit if not expected number

	LDA	tcp_flag			; get TCP flags
	AND	#$04				; check RST bit
	BEQ	LAB_22330			; branch if not reset

	LDX	#$01				; else reset connection
	STX	tcp_state			; TCP state to listen

	DEX					; clear X
	STX	tcp_toc			; clear TCP timeout count
	RTS

; was not reset so get data

LAB_22330
	LDA	tcp_flag			; get TCP flags
	AND	#$02				; check SYN bit
	BNE	LAB_22298			; ignore SYN (should send reset)

	LDA	tcp_flag			; get TCP flags
	AND	#$10				; check ACK bit
	BEQ	LAB_22298			; ignore if no ACK

	JSR	LAB_22815			; prepare TCP reply

						; set tp to TCP data start
	CLC					; clear carry for add
	LDA	tcp_datap_l			; get TCP data pointer low byte
	ADC	#<pk_buff			; add buffer pointer low byte
	STA	tp_l				; save to temp pointer low byte
	LDA	tcp_datap_h			; get TCP data pointer high byte
	ADC	#>pk_buff			; add buffer pointer high byte
	STA	tp_h				; save to temp pointer high byte

						; calculate end pointer
	LDA	tl_l				; get IP length low byte
	ADC	#<ip_buff			; add IP buffer pointer low byte
	STA	te_l				; save to end pointer low byte
	LDA	tl_h				; get IP length high byte
	ADC	#>ip_buff			; add IP buffer pointer high byte
	STA	te_h				; save to end pointer high byte

	LDY	#$01				; set to end +1
	LDA	#$00				; clear A
	STA	(te_l),Y			; null terminate buffer

						; now process the data from the block
LAB_22345
	LDX	#$FF				; set so first increment clears index
	LDY	#$FF				; set so first increment clears index
LAB_22350
	INX					; increment pointer ..
	LDA	LAB_22399a,X		; get vector low byte
	STA	com_vec_l			; save vector low byte
	LDA	LAB_22399,X			; get string pointer low byte
	STA	com_str_l			; save string pointer low byte
	INX					; .. to next word
	LDA	LAB_22399a,X		; get vector high byte
	STA	com_vec_h			; save vector high byte
	LDA	LAB_22399,X			; get string pointer high byte
	BEQ	LAB_ccommexit		; was last vector so go execute it

	STA	com_str_h			; save string pointer high byte
LAB_comparecomm
	INY					; increment index
	LDA	(com_str_l),Y		; get next byte of command to test
	BEQ	LAB_ccommexitt		; exit if end of command (match)

	EOR	(tp_l),Y			; compare with next byte of command to find
	BEQ	LAB_comparecomm		; loop if character match

	BNE	LAB_22350			; branch if not this command

LAB_ccommexit
	INY					; increment index past last byte
LAB_ccommexitt
	CLC					; clear carry for add
	TYA					; copy offset to first byte past command
	ADC	tp_l				; add temp pointer low byte
	STA	tp_l				; save to temp pointer low byte
	LDA	#$00				; set for add carry
	ADC	tp_h				; add temp pointer high byte
	STA	tp_h				; save to temp pointer high byte
	JSR	LAB_vec_comm		; go do command vector
						; (effectively JSR (vec))
						; RTS from the routine will get back to here ..

	LDA	tp_l				; get pointer low byte
	EOR	te_l				; compare with end pointer low byte
	BNE	LAB_22345			; loop if not at end

	LDA	tp_h				; get pointer high byte
	EOR	te_h				; compare with end pointer high byte
	BNE	LAB_22345			; loop if not at end

	JSR	LAB_22925			; generate ACK for data

	LDA	tcp_buff+TCP_flags	; get flags
	ORA	#$01				; flag FIN
	STA	tcp_buff+TCP_flags	; save flags

	LDA	tcp_ds_l			; get tcp_ds low byte
	BNE	LAB_22390			; branch if no rollunder

	DEC	tcp_ds_h			; else decrement high byte
LAB_22390
	DEC	tcp_ds_l			; decrement low byte

	JSR	LAB_22845			; finalise TCP reply
	JSR	LAB_21800			; generate IP header
	JSR	LAB_21200			; transmit packet

	LDA	#$04				; TCP state to fin wait 1
	STA	tcp_state			; save TCP state

	LDA	#$C8				; set TCP timeout count
	STA	tcp_toc			; save TCP timeout count
LAB_22398
	RTS

LAB_vec_comm
	JMP	(com_vec_l)			; do vector


; commands we accept

LAB_22399
	.word	LAB_GET			; "GET"
	.word	LAB_CRLF			; [EOL]
	.word	$0000				; end marker

LAB_GET
	.byte	"GET ",$00			; 
LAB_CRLF
	.byte	$0D,$0A,$00			; [EOL]

LAB_22399a
	.word	LAB_comm_GET		; do GET
	.word	LAB_comm_CRLF		; do [EOL]
	.word	LAB_dumpline		; discard to [EOL] or [EOF]

; null terminate the GET/HEAD command line

LAB_getline
	LDY	#$00				; clear index
LAB_getloop
	LDA	(tp_l),Y			; get byte
	BEQ	LAB_getoutt			; branch if end of buffer

	LDX	#LAB_ell-LAB_el-$01	; byte check table length
LAB_gl_byte
	CMP	LAB_el,X			; compare with table byte
	BEQ	LAB_getout			; branch if end of string

	DEX					; decrement table index
	BPL	LAB_gl_byte			; loop if not all done

	INY					; increment index
	BNE	LAB_getloop			; loop if no rollover

	INC	tp_h				; else increment high byte
	BNE	LAB_getloop			; loop (branch always)
	
LAB_getout
	LDA	#$00				; clear A
	STA	(tp_l),Y			; clear byte to terminate string
LAB_getoutt
	RTS

LAB_el
	.byte	" ?&;"			; characters that can end GET/HEAD strings
LAB_ell


; do HTTP GET

LAB_comm_GET
	LDA	tp_l				; get pointer low byte
	STA	findname_l			; save to find name pointer low byte
	LDA	tp_h				; get pointer high byte
	STA	findname_h			; save to find name pointer high byte
	JSR	LAB_getline			; null terminate the command line
	JSR	LAB_22700			; generate HTTP response


; just dump the lot as a blank line indicates the end of the request

LAB_comm_CRLF
	LDA	te_l				; get end pointer low byte
	STA	tp_l				; save pointer low byte
	LDA	te_h				; get end pointer high byte
	STA	tp_h				; save pointer high byte
	RTS


; dump the rest of the line up to the next [CR][LF]

LAB_dumpline
	LDY	#$00				; clear index
LAB_dumploop
	LDA	(tp_l),Y			; get byte
	BEQ	LAB_dumpout			; branch if end of buffer

	EOR	#$0D				; compare with [CR]
	BNE	LAB_dumpinc			; branch if not [CR]

	INY					; increment index
	BNE	LAB_dumplf			; skip high increment if no rollover

	INC	tp_h				; else increment high byte
LAB_dumplf
	LDA	(tp_l),Y			; get byte
	BEQ	LAB_dumpout			; branch if end of buffer

	EOR	#$0A				; compare with [LF]
	BNE	LAB_dumpinc			; branch if not [LF]

	SEC					; set carry for add +1
	BCS	LAB_dumpoutt		; exit

LAB_dumpinc
	INY					; increment index
	BNE	LAB_dumploop		; loop if no rollover

	INC	tp_h				; else increment high byte
	BNE	LAB_dumploop		; loop (branch always)
	
LAB_dumpout
	CLC					; clear carry for add
LAB_dumpoutt
	TYA					; copy offset
	ADC	tp_l				; add temp pointer low byte
	STA	tp_l				; save temp pointer low byte
	LDA	#$00				; clear for add carry
	ADC	tp_h				; add temp pointer high byte
	STA	tp_h				; save temp pointer high byte
	RTS


; *********************************************************************
; TCP FIN wait 1 state

LAB_22400
	LDA	tcp_toc			; get TCP count
	BNE	LAB_22405			; branch if not timed out

	LDA	#$01				; TCP state to listen
	STA	tcp_state			; save TCP state
	RTS

LAB_22405
	LDX	#$03				; set index
LAB_22410
	LDA	ip_recd,X			; get TCP sender's IP address byte
	CMP	ip_buff+IP_src_add,X	; compare with received IP address
	BNE	LAB_22498			; exit if not expected address

	DEX					; decrement index
	BPL	LAB_22410			; loop if not all done

	LDA	src_port_l			; get sender's port low byte
	CMP	exp_port_l			; compare with expected port low byte
	BNE	LAB_22498			; exit if wrong port

	LDA	src_port_h			; get sender's port high byte
	CMP	exp_port_h			; compare with expected port high byte
	BNE	LAB_22498			; exit if wrong port

	JSR	LAB_22800			; ck rx seq#
	BNE	LAB_22498			; exit if not expected #

	LDA	tcp_flag			; get TCP flags
	AND	#$04				; check RST bit
	BEQ	LAB_22430			; branch if not reset

	LDX	#$01				; else reset connection
	STX	tcp_state			; TCP state to listen
	DEX					; clear X
	STX	tcp_toc			; clear TCP timeout count
	RTS

LAB_22430
	LDA	tcp_flag			; get TCP flags
	AND	#$02				; check SYN bit
	BNE	LAB_22498			; ignore SYN (should send reset)

	LDA	tcp_flag			; get TCP flags
	AND	#$10				; check ACK bit
	BEQ	LAB_22498			; ignore no ACK

	LDA	tcp_flag			; get TCP flags
	AND	#$01				; check FIN bit
	BEQ	LAB_22445			; branch if not FIN

	JMP	LAB_22545			; ACK and close if FIN

LAB_22445
	LDA	#$05				; TCP state to fin wait 2
	STA	tcp_state			; save TCP state
	LDA	#$C8				; set TCP timeout count
	STA	tcp_toc			; save TCP timeout count
LAB_22498
	RTS


; *********************************************************************
; TCP FIN wait 2 state

LAB_22500
	LDA	tcp_toc			; get TCP count
	BNE	LAB_22505			; branch if not timed out

	LDA	#$01				; TCP state to listen
	STA	tcp_state			; save TCP state
	RTS

LAB_22505
	LDX	#$03				; set index
LAB_22510
	LDA	ip_recd,X			; get sender's IP address byte
	CMP	ip_buff+IP_src_add,X	; compare with received IP address
	BNE	LAB_22598			; exit if not expected address

	DEX					; decrement index
	BPL	LAB_22510			; loop if not all done

	LDA	src_port_l			; get sender's port low byte
	CMP	exp_port_l			; compare with expected port low byte
	BNE	LAB_22598			; exit if wrong port

	LDA	src_port_h			; get sender's port high byte
	CMP	exp_port_h			; compare with expected port high byte
	BNE	LAB_22598			; exit if wrong port

	JSR	LAB_22800			; ck rx seq#
	BNE	LAB_22598			; exit if not expected #

	LDA	tcp_flag			; get TCP flags
	AND	#$04				; check RST bit
	BEQ	LAB_22530			; branch if not reset

	LDX	#$01				; else reset connection
	STX	tcp_state			; TCP state to listen
	DEX					; clear X
	STX	tcp_toc			; clear TCP timeout count
	RTS

LAB_22530
	LDA	tcp_flag			; get TCP flags
	AND	#$02				; check SYN bit
	BNE	LAB_22598			; ignore SYN (should send reset)

	LDA	tcp_flag			; get TCP flags
	AND	#$10				; check ACK bit
	BEQ	LAB_22598			; ignore no ACK

	LDA	tcp_flag			; get TCP flags
	AND	#$01				; check FIN bit
	BEQ	LAB_22598			; exit if not FIN

; now ACK the FIN and close

LAB_22545
	JSR	LAB_22815			; set data pointer & prepare TCP reply
	JSR	LAB_22920			; generate ACK for FIN
	JSR	LAB_22845			; finalise TCP reply
	JSR	LAB_21800			; generate IP header
	JSR	LAB_21200			; transmit packet

	LDX	#$01				; set TCP state to listen
	STX	tcp_state			; save TCP state

	DEX					; clear X
	STX	tcp_toc			; clear TCP timeout count
LAB_22598
	RTS


; *********************************************************************
; generate HTTP response

LAB_22700
	JSR	LAB_find			; go find file
	BCS	LAB_isfound			; branch if file exists

	LDA	e_404+$01			; get "error 404" length low byte
	STA	lensys_l			; save length low byte
	LDA	e_404+$02			; get "error 404" length high byte
	STA	lensys_h			; save length high byte

	LDA	#<e_404_b			; get "error 404" pointer low byte
	STA	filesys_l			; save as file system pointer low byte
	LDA	#>e_404_b			; get "error 404" pointer high byte
	STA	filesys_h			; save as file system pointer high byte

LAB_isfound
	CLC					; clear carry for add
	LDA	#<pk_buff			; get buffer pointer low byte
	ADC	tx_len_l			; add tx length (next free) low byte
	STA	file_l			; set pointer low byte
	LDA	#>pk_buff			; get buffer pointer high byte
	ADC	tx_len_h			; add tx length (next free) high byte
	STA	file_h			; set pointer high byte

						; now copy the file into the buffer
	LDY	#$00				; clear index
LAB_isfound_lp
	LDA	lensys_l			; get length low byte
	BNE	LAB_isfound_de		; branch if no underflow

	DEC	lensys_h			; decrement length high byte
	BMI	LAB_22770			; exit if done

LAB_isfound_de
	DEC	lensys_l			; decrement length low byte

	LDA	(filesys_l),Y		; get byte from file
	STA	(file_l),Y			; save byte to packet
	INY					; increment index
	BNE	LAB_no_sysh_inc		; branch if no rollover

	INC	filesys_h			; increment source pointer high byte
	INC	file_h			; increment destination pointer high byte
LAB_no_sysh_inc
	INC	tx_len_l			; increment tx length low byte
	BNE	LAB_isfound_lp		; loop if no rollover

	INC	tx_len_h			; increment tx length high byte
	BNE	LAB_isfound_lp		; loop (branch always)

LAB_22770
	RTS


; *********************************************************************
; check TCP sequence, return Zb=1 if match

LAB_22800
	LDX	#$03				; set index
LAB_22805
	LDA	tcp_buff+TCP_seq_n,X	; get received sequence byte
	EOR	seq_recd,X			; compare with next rx sequence byte
	BNE	LAB_22810			; branch if no match

	DEX					; decrement index
	BPL	LAB_22805			; loop if not all done

	INX					; set Zb
LAB_22810
	RTS


; *********************************************************************
; prepare TCP reply. copy our sequence number to header, set header
; length in header, clear flags and urgent pointer in header, set TCP
; data start and Tx total length.

LAB_22815
	LDX	#$03				; set index
LAB_22816
	LDA	seq_ours,X			; get my sequence #
	STA	tcp_buff+TCP_seq_n,X	; save sequence number
	DEX					; decrement index
	BPL	LAB_22816			; loop if not all done


	LDA	#$50				; set TCP header length
	STA	tcp_buff+TCP_d_oset	; save TCP header length

	LDA	#$36				; set data start low byte
	STA	tcp_ds_l			; save data start low byte
	STA	tx_len_l			; save total length low byte

	INX					; clear X
	STX	tcp_ds_h			; clear data start high byte
	STX	tx_len_h			; clear total length high byte

	STX	tcp_buff+TCP_flags	; clear flags
	STX	tcp_buff+TCP_urg_p_h	; clear urgent pointer high byte
	STX	tcp_buff+TCP_urg_p_l	; clear urgent pointer low byte

	RTS


; *********************************************************************
; generate SYN

LAB_22895
	LDA	tcp_buff+TCP_flags	; get flag byte
	ORA	#$02				; set SYN flag
	STA	tcp_buff+TCP_flags	; save flag byte

	CLC					; clear carry for add
	LDA	tcp_buff+TCP_d_oset	; get TCP header length/type byte
	ADC	#$10				; add TCP header length for option
	STA	tcp_buff+TCP_d_oset	; save TCP header length/type byte

	LDX	tx_len_l			; get Tx length low byte (length is < $100)
	LDA	#$02				; opt 2
	STA	pk_buff,X			; save option
	INX					; increment length/pointer
	LDA	#$04				; length 4
	STA	pk_buff,X			; save length
	INX					; increment length/pointer
	LDA	#$02				;.
	STA	pk_buff,X			;.
	INX					; increment length/pointer
	LDA	#$00				;.
	STA	pk_buff,X			;.
	STA	tcp_ds_h			; clear high byte
	STX	tcp_ds_l			; tcp_ds = tx_len-1, sequence space for our SYN
	INX					; increment length/pointer
	STX	tx_len_l			; update length
	RTS


; *********************************************************************
; generate ACK for SYN or FIN received

LAB_22920
	SEC					; +1 for SYN or FIN
	BCS	LAB_22926			; go do ACK

; generate ACK for TCP data

LAB_22925
	CLC					; clear carry for add
LAB_22926
	LDA	seq_recd+$03		; get received sequence no low byte
	ADC	tcp_d_len_l			; add TCP data size low byte
	STA	seq_recd+$03		; save received sequence no low byte

	LDA	seq_recd+$02		; get received sequence no byte
	ADC	tcp_d_len_h			; add TCP data size high byte
	STA	seq_recd+$02		; save received sequence no byte

	LDA	seq_recd+$01		; get received sequence no byte
	ADC	#$00				; add carry
	STA	seq_recd+$01		; save received sequence no byte

	LDA	seq_recd			; get received sequence no high byte
	ADC	#$00				; add carry
	STA	seq_recd			; save received sequence no high byte

	LDX	#$03
LAB_22927
	LDA	seq_recd,X			; get rx sequence no
	STA	tcp_buff+TCP_ack_n,X	; save as ack no
	DEX					; decrement index
	BPL	LAB_22927			; loop if not all done

	LDA	tcp_buff+TCP_flags	; get flags
	ORA	#$10				; flag ACK
	STA	tcp_buff+TCP_flags	; save flags

						; this bites! do it properly!!
	LDA	#$05				; set window high byte
	STA	tcp_buff+TCP_wind_h	; save window high byte

	LDA	#$00				; set window low byte
	STA	tcp_buff+TCP_wind_l	; save window low byte

	RTS


; *********************************************************************
; finalise TCP reply

LAB_22845
	LDA	tcp_buff+TCP_src_p_h	; swap port high bytes
	LDX	tcp_buff+TCP_dst_p_h	;
	STA	tcp_buff+TCP_dst_p_h	;
	STX	tcp_buff+TCP_src_p_h	;
	LDA	tcp_buff+TCP_src_p_l	; swap port low bytes
	LDX	tcp_buff+TCP_dst_p_l	;
	STA	tcp_buff+TCP_dst_p_l	;
	STX	tcp_buff+TCP_src_p_l	;

						; calculate IP payload length
	SEC					; set carry for subtract
	LDA	tx_len_l			; get Tx length low byte
	SBC	#<ip_ofs			; minus header length
	STA	ip_buff+IP_length_l	; save low byte to header
	LDA	tx_len_h			; get Tx length high byte
	SBC	#>ip_ofs			; minus carry
	STA	ip_buff+IP_length_h	; save high byte to header

	LDA	#$00				; clear A
	STA	tcp_buff+TCP_cksum_h	; clear checksum high byte
	STA	tcp_buff+TCP_cksum_l	; clear checksum low byte

	JSR	LAB_21900			; do TCP checksum

	LDA	chksum_h			; get checksum high byte
	EOR	#$FF				; ones complement
	STA	tcp_buff+TCP_cksum_h	; save checksum high byte
	LDA	chksum_l			; get checksum low byte
	EOR	#$FF				; ones complement
	STA	tcp_buff+TCP_cksum_l	; save checksum low byte

	LDA	#$3C				; set time to live
	STA	ip_buff+IP_t2l		; save time to live


; *********************************************************************
; generate our next sequence number

LAB_22975
	SEC					; set carry for subtract
	LDA	tx_len_l			; get length low byte
	SBC	tcp_ds_l			; subtract data start low byte
	STA	aa_l				; save TCP data size low byte
	LDA	tx_len_h			; get length high byte
	SBC	tcp_ds_h			; subtract data start high byte
	STA	aa_h				; save TCP data size high byte

	CLC					; clear carry for add
	LDA	seq_ours+$03		; get our sequence no low byte
	ADC	aa_l				; add TCP data size low byte
	STA	seq_ours+$03		; save our sequence no low byte

	LDA	seq_ours+$02		; get our sequence no byte
	ADC	aa_h				; add TCP data size high byte
	STA	seq_ours+$02		; save our sequence no byte

	LDA	seq_ours+$01		; get our sequence no byte
	ADC	#$00				; add carry
	STA	seq_ours+$01		; save our sequence no byte

	LDA	seq_ours			; get our sequence no high byte
	ADC	#$00				; add carry
	STA	seq_ours			; save our sequence no high byte
	RTS


; *********************************************************************
; generate 8 bit PRNG

LAB_get_prng
	LDA	prng			; get byte
	ASL				; *2
	ASL				; *4
	SEC				; set for add + 1
	ADC	prng			; *5 + 1
	STA	prng			; save result
	RTS


; *********************************************************************
; find file, searches the file system for the filename pointed to by
; the find name pointer. returns with Cb = 1 and the pointer to the
; file payload in filesys, if found.
;
; file name to search for should be "/[path/[path/]][name]" where path is
; the directory name followed by a "/". the path may be many deep and
; both the path and the name may be null. if the name is null then the
; default_file name will be searched for the path given.

LAB_find
	LDA	root+$01			; get root file payload length low byte
	STA	lensys_l			; save payload length low byte
	LDA	root+$02			; get root file payload length high byte
	STA	lensys_h			; save payload length high byte

	LDA	#<root_b			; get root body pointer low byte
	STA	filesys_l			; set file system pointer low byte
	LDA	#>root_b			; get root body pointer high byte
	STA	filesys_h			; set file system pointer high byte
	LDY	#$00				; clear index
	LDA	(findname_l),Y		; get the first byte of the name to find
	CMP	#'/'				; compare with separator
	BNE	LAB_exit_n_found		; exit if not "/" at start

	JSR	LAB_directory		; search the root directory for the file
	BVC	LAB_exit_find		; exit if it's a file

	JSR	LAB_find_default		; it's a directory go find the default file
LAB_exit_find
	RTS


; flag file found and exit

LAB_exit_found
	BIT	fileflags			; test flags byte, set Vb for a directory
	SEC					; flag found
	RTS


; flag file not found and exit

LAB_exit_n_found
	CLV					; flag not a directory
	CLC					; flag not found
	RTS


; increment filesys to the next file pointer in the directory

LAB_nextfile
	CLC					; clear carry for add
	LDA	filesys_l			; get filesys low byte
	ADC	#$02				; increment to next pointer
	STA	filesys_l			; save filesys low byte
	BCC	ni_inc_h			; branch if no rollover

	INC	filesys_h			; else increment filesys high byte
ni_inc_h
	SEC					; set carry for subtract
	LDA	lensys_l			; get remaining length low byte
	SBC	#$02				; increment to next pointer
	STA	lensys_l			; save remaining length low byte
	BCS	LAB_comparefile		; branch if no rollunder

	DEC	lensys_h			; decerment remaining length high byte

LAB_comparefile
	LDA	lensys_l			; get remaining directory length low byte
	ORA	lensys_h			; OR remaining directory length high byte
	BEQ	LAB_exit_n_found		; exit if no more directory entries

	LDY	#$00				; clear index
	LDA	(filesys_l),Y		; get file pointer low byte
	STA	file_l			; save this file pointer low byte
	INY					; increment index
	LDA	(filesys_l),Y		; get file pointer high byte
	STA	file_h			; save this file pointer high byte

	DEY					; clear index
	LDA	(file_l),Y			; get this file's flags
	STA	fileflags			; save the file's flag byte
	INY					; point to payload length low byte
	LDA	(file_l),Y			; get this file's payload length low byte
	STA	length_l			; save this file's payload length low byte
	INY					; point to payload length high byte
	LDA	(file_l),Y			; get this file's payload length high byte
	STA	length_h			; save this file's payload length high byte

	CLC					; clear carry for add
	LDA	file_l			; get this file pointer low byte
	ADC	#f_name-f_start		; add offset to the file name
	STA	file_l			; save this file pointer low byte
	BCC	nf_inc_h			; branch if no rollover

	INC	file_h			; else increment this file pointer high byte
nf_inc_h
	LDY	#$FF				; set so first increment clears index

; compare this file's name, pointed to by file, with the find file name
; pointed to by findname. ends with Y indexed to the next byte in the
; name if the whole name matched.

LAB_comparename
	INY					; increment index
	LDA	(file_l),Y			; get next byte of name to test
	BEQ	LAB_cnameexit		; exit if end of name (match)

	EOR	(findname_l),Y		; compare with next byte of name to find
	BEQ	LAB_comparename		; loop if character match

	BNE	LAB_nextfile		; branch if not this file

LAB_cnameexit
	LDA	(findname_l),Y		; get next byte of name to find
	BEQ	LAB_end_find		; branch if end of name to find

	CMP	#'/'				; compare with separator
	BNE	LAB_nextfile		; branch if not end of name to find

LAB_end_find
	PHA					; save next byte of name to find

						; name matched so update lensys
	LDA	length_l			; get length low byte
	STA	lensys_l			; save length low byte
	LDA	length_h			; get length high byte
	STA	lensys_h			; save length high byte

						; name matched so update filesys
	SEC					; set carry for add +1, past $00
	TYA					; copy offset
	ADC	file_l			; add this file pointer low byte
	STA	filesys_l			; save as file system pointer low byte
	LDA	file_h			; get this file pointer high byte
	ADC	#$00				; add in the carry
	STA	filesys_h			; save as file system pointer high byte

	PLA					; restore next byte of name to find
	BEQ	LAB_exit_found		; branch if end of name to find

	BIT	fileflags			; else test the flags byte
	BVC	LAB_exit_n_found		; branch if not a directory


; searches the directory pointed to by the file system pointer for the
; filename pointed to by the find name pointer. the routine is entered
; with filesys pointing to the first directory byte after the file header
; and lensys holding the remaining directory size

LAB_directory				; ok so far so update findname
	SEC					; set carry for add +1, past "/" or $00
	TYA					; copy offset
	LDY	#$00				; clear index
	ADC	findname_l			; add offset to findname low byte
	STA	findname_l			; save findname low byte
	TYA					; set $00 for add carry
	ADC	findname_h			; add findname high byte
	STA	findname_h			; save findname high byte

	LDA	(findname_l),Y		; get next byte of name to find
	BNE	LAB_comparefile		; go compare file if not null filename

LAB_find_default				; else get default file for this directory
	LDA	#<default_file		; get default filename pointer low byte
	STA	findname_l			; set pointer of name to find low byte
	LDA	#>default_file		; get default filename pointer high byte
	STA	findname_h			; set pointer of name to find high byte
	BNE	LAB_comparefile		; go compare file (branch always)


; *********************************************************************
; this is the default file in all directories. this will be the target
; if the path ends in "/" and can be a unique file in each directory

default_file
	.byte	"index.html",$00		; 


; *********************************************************************
; this is the test data for the filesystem. it is made up of the HTML
; pages and other files for the Vic 20 mini web server

; structure for this trial is ..

; root+--errors+--index.html		; really error 403 - Forbidden
;	|	   +--404			; error 404 - Not found
;	|
;	+--images+--back.png		; background .png image
;	|	   +--e.gif			; small .gif image
;	|	   +--index.html		; really error 403 - Forbidden
;	|	   +--up.gif		; small .gif image
;	|
;	+--b.html				; button page
;	+--index.html			; home page
;	+--o.html				; other page


; *********************************************************************
; the root file should always be a directory

root
	.byte	directory			; flags
	.word	root_end-root_b		; file payload length
	.byte	$00				; filename, null in this case
root_b
	.word	errors			; errors directory
	.word	images			; images directory
						; done directories, now do files
	.word	b_html			; the button page
	.word	index_html			; default web page
	.word	o_html			; other web page
root_end


; *********************************************************************
; home page, served for "GET / " and "GET /index.html"

index_html
	.byte	none				; flags
	.word	index_end-index_html_b	; file payload length
	.byte	"index.html",$00		; filename, null terminator
						; rest is file payload
index_html_b
	.byte	"HTTP/1.0 200 OK",$0D,$0A,"Content-Type: text/html"
	.byte	$0D,$0A,$0D,$0A
	.byte	"<HTML><HEAD><TITLE>Vic 20 WEB Server</TITLE></HEAD><BODY BACK"
	.byte	"GROUND=images/back.png><TABLE WIDTH=100%><TR><TD><IMG SRC=ima"
	.byte	"ges/up.gif BORDER=0 ALT=Up></TD><TD ALIGN=CENTER><FONT SIZE=+"
	.byte	"1><B>The home page </FONT></B><FONT SIZE=-1>By Lee Davison</F"
	.byte	"ONT></TD><TD><IMG SRC=images/up.gif ALIGN=RIGHT BORDER=0 ALT="
	.byte	"Top></TD></TR></TABLE><HR><TABLE ALIGN=CENTER WIDTH=80% HEIGH"
	.byte	"T=70%><TR><TD VALIGN=CENTER><FONT SIZE=+1>Vic 20 WEB Server <"
	.byte	"/FONT>- <A HREF=b.html>The button page</A> - <A HREF=o.html>T"
	.byte	"he other page</A><P><FONT SIZE=+4><B>Welcome </B></FONT><FONT"
	.byte	" FACE=Courier>to the home page</FONT><P><A HREF=mailto:leeeda"
	.byte	"vison@lycos.co.uk>&nbsp;e-mail me <IMG SRC=images/e.gif BORDE"
	.byte	"R=0 ALT=e-mail></A></TD></TR></TABLE><HR><FONT SIZE=-1>Last p"
	.byte	"age update: 29th February, 2004</FONT></BODY></HTML>"
index_end
 
; *********************************************************************
; other web page, served for "GET /o.html"

o_html
	.byte	none				; flags
	.word	o_html_end-o_html_b	; file payload length
	.byte	"o.html",$00		; filename, null terminator
						; rest is file payload
o_html_b
	.byte	"HTTP/1.0 200 OK",$0D,$0A,"Content-Type: text/html"
	.byte	$0D,$0A,$0D,$0A
	.byte	"<HTML><HEAD><TITLE>Vic 20 WEB Server</TITLE></HEAD><BODY BACK"
	.byte	"GROUND=images/back.png><TABLE WIDTH=100%><TR><TD><IMG SRC=ima"
	.byte	"ges/up.gif BORDER=0 ALT=Up></TD><TD ALIGN=CENTER><FONT SIZE=+"
	.byte	"1><B>The other page </B></FONT><FONT SIZE=-1>By Lee Davison</"
	.byte	"FONT></TD><TD><IMG SRC=images/up.gif ALIGN=RIGHT BORDER=0 ALT"
	.byte	"=Top></TD></TR></TABLE><HR><TABLE ALIGN=CENTER WIDTH=80% HEIG"
	.byte	"HT=70%><TR><TD VALIGN=CENTER><FONT SIZE=+1>Vic 20 WEB Server "
	.byte	"</FONT>- <A HREF=index.html>The home page</A> - <A HREF=b.htm"
	.byte	"l>The button page</A><P><FONT SIZE=+4><B>Welcome </B></FONT><"
	.byte	"FONT FACE=Courier>to the other page</FONT><P><A HREF=mailto:l"
	.byte	"eeedavison@lycos.co.uk>&nbsp;e-mail me <IMG SRC=images/e.gif "
	.byte	"BORDER=0 ALT=e-mail></A></TD></TR></TABLE><HR><FONT SIZE=-1>L"
	.byte	"ast page update: 29th February, 2004</FONT></BODY></HTML>"
o_html_end


; *********************************************************************
; the button page, served for "GET /b.html"

b_html
	.byte	none				; flags
	.word	b_html_end-b_html_b	; file payload length
	.byte	"b.html",$00		; filename, null terminator
						; rest is file payload
b_html_b
	.byte	"HTTP/1.0 200 OK",$0D,$0A,"Content-Type: text/html"
	.byte	$0D,$0A,$0D,$0A
	.byte	"<HTML><HEAD><TITLE>Vic 20 WEB Server</TITLE></HEAD><BODY BACK"
	.byte	"GROUND=images/back.png><TABLE WIDTH=100%><TR><TD><IMG SRC=ima"
	.byte	"ges/up.gif BORDER=0 ALT=Up></TD><TD ALIGN=CENTER><FONT SIZE=+"
	.byte	"1><B>The button page </B></FONT><FONT SIZE=-1>By Lee Davison<"
	.byte	"/FONT></TD><TD><IMG SRC=images/up.gif ALIGN=RIGHT BORDER=0 AL"
	.byte	"T=Top></TD></TR></TABLE><HR><TABLE ALIGN=CENTER WIDTH=80% HEI"
	.byte	"GHT=70%><TR><TD VALIGN=CENTER><FONT SIZE=+1>Vic 20 WEB Server"
	.byte	" </FONT>- <A HREF=index.html>The home page</A> - <A HREF=o.ht"
	.byte	"ml>The other page</A><P><FONT SIZE=+4><B>Welcome </B></FONT><"
	.byte	"FONT FACE=Courier>to the button page</FONT><P><A HREF=mailto:"
	.byte	"leeedavison@lycos.co.uk>&nbsp;e-mail me <IMG SRC=images/e.gif"
	.byte	" BORDER=0 ALT=e-mail></A></TD><TD WIDTH=20% ALIGN=CENTER>Are "
	.byte	"you easily confused?<FORM METHOD=GET ACTION=b.html><PRE>yes <"
	.byte	"INPUT TYPE=RADIO NAME=4 VALUE=0>",$0D,$0A,"no  <INPUT TYPE=RA"
	.byte	"DIO NAME=4 VALUE=1>",$0D,$0A,"fish<INPUT TYPE=RADIO NAME=4 VA"
	.byte	"LUE=2 CHECKED>",$0D,$0A,"<INPUT TYPE=SUBMIT VALUE=Submit></PR"
	.byte	"E></FORM></TD></TR></TABLE><HR><FONT SIZE=-1>Last page update"
	.byte	": 29th February, 2004</FONT></BODY></HTML>"
b_html_end


; *********************************************************************
; errors directory

errors
	.byte	directory			; flags
	.word	errors_end-errors_b	; file payload length
	.byte	"errors",$00		; filename, null terminator
errors_b
	.word	e_deflt			; default file for errors directory
	.word	e_404				; error 404, served for not found files
errors_end


; default file for errors directory

e_deflt
	.byte	none				; flags
	.word	e_deflt_end-e_deflt_b	; file payload length
	.byte	"index.html",$00		; filename, null terminator
						; rest is file payload
e_deflt_b
	.byte	"HTTP/1.0 403 Forbidden",$0D,$0A,"Content-Type: text/plain"
	.byte	$0D,$0A,$0D,$0A
	.byte	"<H1>Forbidden</H1>"
e_deflt_end

; error 404, served for not found files

e_404
	.byte	none				; flags
	.word	e_404_end-e_404_b		; file payload length
	.byte	"404",$00			; filename, null terminator
						; rest is file payload
e_404_b
	.byte	"HTTP/1.0 404 Not Found",$0D,$0A,"Content-Type: text/plain"
	.byte	$0D,$0A,$0D,$0A
	.byte	"<H1>File not found</H1>"
e_404_end


; *********************************************************************
; images directory

images
	.byte	directory			; flags
	.word	images_end-images_b	; file payload length
	.byte	"images",$00		; filename, null terminator
images_b
	.word	b_png				; background image
	.word	e_gif				; small image
	.word	i_indx			; default page
	.word	up_gif			; small image
images_end


; background image, served for "GET images/back.png"

b_png
	.byte	none				; flags
	.word	b_png_end-b_png_b		; file payload length
	.byte	"back.png",$00		; filename, null terminator
						; rest is file payload
b_png_b
	.byte	"HTTP/1.0 200 OK",$0D,$0A,"Content-type: image/png",$0D,$0A
	.byte	$0D,$0A
	.byte	$89,$50,$4E,$47,$0D,$0A,$1A,$0A,$00,$00,$00,$0D,$49,$48,$44,$52
	.byte	$00,$00,$00,$7F,$00,$00,$00,$A2,$01,$03,$00,$00,$01,$25,$D4,$B0
	.byte	$37,$00,$00,$00,$06,$50,$4C,$54,$45,$FB,$FB,$FB,$DB,$DB,$DB,$46
	.byte	$A0,$6D,$A6,$00,$00,$01,$7A,$49,$44,$41,$54,$78,$9C,$ED,$54,$C1
	.byte	$4A,$C3,$40,$10,$7D,$DB,$36,$54,$50,$4C,$C5,$4B,$0E,$42,$EA,$1F
	.byte	$F4,$28,$52,$6C,$3D,$F9,$1B,$F9,$0E,$29,$4D,$BF,$C0,$3F,$12,$7A
	.byte	$F0,$EC,$37,$E4,$20,$1E,$F4,$12,$F0,$60,$0E,$49,$C6,$D9,$EC,$6E
	.byte	$93,$AC,$D1,$44,$11,$A1,$90,$49,$18,$DE,$BE,$9D,$99,$97,$5D,$26
	.byte	$83,$8D,$00,$5B,$98,$4B,$3F,$9A,$42,$99,$02,$AB,$AB,$16,$92,$08
	.byte	$10,$F2,$BD,$BD,$97,$FE,$E0,$5C,$2F,$15,$48,$16,$4D,$64,$91,$AE
	.byte	$B0,$4C,$2F,$36,$4B,$17,$1C,$69,$14,$46,$6E,$66,$73,$DD,$42,$20
	.byte	$97,$50,$4B,$22,$A5,$31,$AC,$B8,$25,$70,$88,$45,$4A,$6F,$D5,$8D
	.byte	$C0,$33,$6E,$C7,$51,$E4,$D3,$7F,$85,$A4,$A8,$58,$E0,$01,$CB,$C0
	.byte	$1B,$C8,$2F,$35,$21,$F2,$1C,$E6,$30,$D2,$C6,$25,$98,$28,$70,$C3
	.byte	$60,$0E,$11,$31,$00,$E6,$0E,$83,$71,$43,$70,$48,$29,$04,$D1,$7B
	.byte	$C3,$16,$54,$7A,$0D,$EC,$6D,$7A,$93,$B9,$94,$F1,$43,$B1,$B9,$B1
	.byte	$26,$09,$D2,$66,$98,$E3,$4E,$60,$A6,$C0,$63,$09,$80,$0B,$88,$2D
	.byte	$6F,$F9,$F4,$1A,$67,$21,$DD,$31,$73,$09,$67,$5B,$64,$0D,$30,$42
	.byte	$1B,$D0,$C1,$44,$2F,$70,$42,$CA,$FB,$F4,$FD,$4C,$8F,$0D,$03,$13
	.byte	$D3,$02,$74,$30,$B7,$0D,$65,$2A,$5D,$37,$D2,$CC,$6A,$B6,$6F,$1B
	.byte	$52,$77,$71,$DE,$00,$2C,$5B,$FF,$09,$E1,$6F,$2B,$4B,$91,$AE,$F9
	.byte	$3C,$FC,$37,$49,$52,$C8,$23,$4C,$D7,$C5,$20,$73,$23,$76,$C3,$84
	.byte	$5D,$A0,$6A,$B8,$4F,$B8,$C6,$69,$52,$16,$1D,$3F,$E0,$0C,$27,$71
	.byte	$45,$45,$5E,$CB,$04,$3F,$21,$B8,$06,$76,$35,$36,$4A,$05,$5A,$25
	.byte	$37,$B2,$B0,$65,$D1,$CB,$F6,$B2,$BD,$6C,$9B,$EC,$B3,$2D,$0B,$3B
	.byte	$05,$76,$D1,$AE,$C4,$A7,$1A,$35,$95,$FA,$FC,$40,$6C,$4D,$18,$AC
	.byte	$EA,$33,$E8,$97,$73,$6C,$53,$27,$76,$63,$93,$EC,$39,$FA,$35,$F1
	.byte	$01,$FD,$CD,$7D,$2F,$89,$09,$9D,$B7,$00,$00,$00,$00,$49,$45,$4E
	.byte	$44,$AE,$42,$60,$82
b_png_end


; small image, served for "GET images/e.gif"

e_gif
	.byte	none				; flags
	.word	e_gif_end-e_gif_b		; file payload length
	.byte	"e.gif",$00			; filename, null terminator
						; rest is file payload
e_gif_b
	.byte	"HTTP/1.0 200 OK",$0D,$0A,"Content-Type: image/gif",$0D,$0A
	.byte	$0D,$0A
	.byte	$47,$49,$46,$38,$39,$61,$10,$00,$10,$00,$C4,$00,$00,$F7,$00,$5F
	.byte	$33,$03,$9A,$33,$03,$CA,$03,$03,$83,$5F,$5F,$F7,$33,$63,$FA,$63
	.byte	$9B,$FB,$63,$CB,$FB,$03,$9B,$CB,$63,$FA,$FA,$FA,$FA,$CB,$97,$00
	.byte	$00,$CB,$03,$03,$FA,$FA,$FA,$63,$63,$63,$03,$03,$03,$FF,$FF,$FF
	.byte	$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00
	.byte	$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00
	.byte	$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$21,$F9,$04
	.byte	$01,$00,$00,$10,$00,$2C,$00,$00,$00,$00,$10,$00,$10,$00,$00,$05
	.byte	$69,$20,$24,$8E,$C6,$91,$24,$63,$9A,$96,$49,$11,$18,$AA,$28,$40
	.byte	$85,$59,$40,$46,$A0,$3A,$8E,$20,$20,$07,$01,$6F,$38,$7C,$38,$1A
	.byte	$3F,$84,$40,$C1,$6C,$36,$8D,$C8,$02,$62,$E0,$64,$02,$18,$0A,$E8
	.byte	$60,$BB,$AD,$32,$16,$D9,$63,$A3,$51,$2D,$43,$C9,$E5,$EA,$B9,$2C
	.byte	$28,$B8,$A9,$D0,$B4,$42,$70,$82,$3B,$E4,$F3,$7A,$F8,$C1,$EF,$F3
	.byte	$05,$26,$5B,$7C,$2A,$2F,$10,$02,$25,$37,$03,$31,$10,$04,$01,$05
	.byte	$06,$06,$37,$8B,$23,$01,$03,$6D,$31,$21,$00,$3B
e_gif_end


; default .html page for "/images/"

i_indx
	.byte	none				; flags
	.word	i_indx_end-i_indx_b	; file payload length
	.byte	"index.html",$00		; filename, null terminator
						; rest is file payload
i_indx_b
	.byte	"HTTP/1.0 403 Forbidden",$0D,$0A,"Content-Type: text/html"
	.byte	$0D,$0A,$0D,$0A
i_indx_end


; small image, served for "GET images/up.gif"

up_gif
	.byte	none				; flags
	.word	up_gif_end-up_gif_b	; file payload length
	.byte	"up.gif",$00		; filename, null terminator
						; rest is file payload
up_gif_b
	.byte	"HTTP/1.0 200 OK",$0D,$0A,"Content-Type: image/gif",$0D,$0A
	.byte	$0D,$0A
	.byte	$47,$49,$46,$38,$39,$61,$40,$00,$14,$00,$91,$00,$00,$00,$00,$FF
	.byte	$BD,$BD,$BD,$00,$00,$00,$FF,$FF,$FF,$21,$F9,$04,$01,$00,$00,$03
	.byte	$00,$2C,$00,$00,$00,$00,$40,$00,$14,$00,$00,$02,$99,$9C,$8F,$A9
	.byte	$CB,$ED,$0F,$A3,$9C,$B4,$DA,$8B,$B3,$86,$61,$E7,$4E,$05,$02,$E8
	.byte	$85,$E3,$24,$0A,$A7,$03,$B4,$C0,$E0,$BE,$57,$BA,$3E,$B4,$4A,$32
	.byte	$ED,$B1,$CF,$EA,$9F,$5B,$DC,$80,$2C,$19,$4C,$16,$EB,$35,$86,$B8
	.byte	$E5,$EF,$D9,$D4,$19,$95,$AE,$A3,$31,$C1,$24,$2A,$B2,$CF,$20,$6F
	.byte	$0A,$FE,$0A,$A1,$E4,$20,$17,$EA,$3D,$8A,$0D,$D4,$AB,$E1,$DC,$3D
	.byte	$C0,$D1,$8A,$F6,$5A,$8D,$98,$C7,$03,$7C,$3E,$19,$D8,$17,$54,$55
	.byte	$C5,$B6,$43,$F8,$16,$F8,$87,$13,$98,$A7,$58,$23,$E5,$96,$E4,$70
	.byte	$96,$36,$40,$79,$A1,$34,$E9,$58,$79,$69,$91,$E9,$A4,$C8,$B9,$E9
	.byte	$19,$63,$33,$3A,$16,$D8,$F7,$58,$82,$35,$92,$2A,$C2,$99,$C6,$C7
	.byte	$CA,$30,$BB,$95,$53,$00,$00,$3B
up_gif_end


; *********************************************************************
; low level routines for stats etc.. most of this is quick & dirty code
; just to get some indication on the screen


; *********************************************************************
; print out our IP address

print_ip
	LDA	#$A1				; cursor low byte
	STA	cur_l				; save it
	LDA	#$1E				; cursor high byte
	STA	cur_h				; save it

	LDX	#$00				; set index
	BEQ	iip_loop			; branch into loop

ip_loop
	LDA	#$2E				; "."
	JSR	chr_out			; go output character
iip_loop
	LDA	ip_ours,X			; get IP address byte
	JSR	print_dec			; print A as decimal byte
	INX					; increment index
	CPX	#$04				; compare with max+1
	BNE	ip_loop			; loop if not finished

	RTS


; *********************************************************************
; print out our MAC address

print_mac
	LDA	#$72				; cursor low byte
	STA	cur_l				; save it
	LDA	#$1E				; cursor high byte
	STA	cur_h				; save it

	LDX	#$00				; set index
mac_loop
	LDA	#$20				; " "
	JSR	chr_out			; go output character
	LDA	mac_ours,X			; get MAC address byte
	JSR	print_hex			; print A as hex byte
	INX					; increment index
	CPX	#$06				; compare with max+1
	BNE	mac_loop			; loop if not finished

	RTS


; *********************************************************************
; print A as hex byte

print_hex
	PHA					; save for now
	LSR					; shift high ..
	LSR					; .. nibble ..
	LSR					; .. to low ..
	LSR					; .. nibble
	CMP	#$0A				; set carry for +1 if >9	
	SED					; set decimal mode
	ADC	#'0'				; add ASCII "0"
	JSR	chr_out			; go output character
	PLA					; restore original
	AND	#$0F				; mask low nibble
	CMP	#$0A				; set carry for +1 if >9	
	ADC	#'0'				; add ASCII "0"
	CLD					; clear decimal mode
	JMP	chr_out			; go output character & return


; *********************************************************************
; print TCP state

print_tcp
	LDA	#$D1				; cursor low byte
	STA	cur_l				; save it
	LDA	#$1E				; cursor high byte
	STA	cur_h				; save it

	LDA	tcp_state			; get TCP state
	ASL					; *2
	ASL					; *4
	ADC	tcp_state			; *5
	ASL					; *10
	TAX					; copy to index
	LDY	#$0A				; set count
p_tcp_loop
	TYA					; copy index
	PHA					; save index
	LDA	tcp_string-$0A,X		; get byte
	JSR	chr_out			; print it
	PLA					; pull index
	TAY					; restore index
	INX					; increment print index
	DEY					; decrement count
	BNE	p_tcp_loop			; loop if not all done

	RTS


; *********************************************************************
; print byte in A as leading zero supressed decimal

print_dec
	SED					; all adds in decimal mode
	STA	b2b_low			; save byte
	TXA					; copy X
	PHA					; save X

	LDA	#$00				; clear 10's & 1's
	TAY					; clear 100's
	LDX	#$07				; set bit count
bit_loop
	LSR	b2b_low			; bit to carry
	BCC	skip_add			; branch if no add

	ADC	b2b_table-$01,X		; else add BCD value
skip_add
	DEX					; decrement bit count
	BNE	bit_loop			; loop if more to do

	BCC	skip_100			; branch if no 100's carry
						; if Cb set here (and can only be set by the last
						; loop add) then there was a carry into the 100's
	INY					; so add 100's carry to the high byte
skip_100
						; now check the 2^7 (128) bit
	LSR	b2b_low			; bit 7 to carry
	BCC	skip_fin			; branch if no add

	INY					; else effectively add 100 part of 128
	ADC	#$27				; and then add 128 (-1) part of 128
	BCC	skip_fin			; branch if no further carry

	INY					; else add 200's carry
	CLC					; clear 100's printed flag
skip_fin
	CLD					; clear decimal mode

						; done converting the byte, now print it
	TAX					; copy result low byte
	TYA					; copy result high byte
	BEQ	skip_p100			; if high byte null then skip 100's print

	ORA	#$30				; convert nibble to ASCII
	JSR	chr_out			; go output 100's character
	SEC					; flag printed 100's
skip_p100
	TXA					; get result low byte
	BCS	pri_10s			; if done 100's then force 10's print

	AND	#$F0				; mask high nibble
	BEQ	pri_1s			; skip 10's print if leading 0

pri_10s
	LSR					; shift high ..
	LSR					; .. nibble ..
	LSR					; .. to low ..
	LSR					; nibble ..
	ORA	#$30				; convert nibble to ASCII
	JSR	chr_out			; go output 10's character
pri_1s
	TXA					; get result low byte (again)
	AND	#$0F				; mask high nibble
	ORA	#$30				; convert nibble to ASCII
	JSR	chr_out			; go output 1's character & return
	PLA					; pull X
	TAX					; restore X
	RTS


; *********************************************************************
; *********************************************************************
; VIC 20 specific code from here on

VIA_1		=	$9110			; VIA 1
VIA_2		=	$9120			; VIA 2

DRB		=	$00			; data register B
DRA		=	$01			; data register A
DDRB		=	$02			; data direction register B
DDRA		=	$03			; data direction register A
T1C_l		=	$04			; timer 1 counter low byte
T1C_h		=	$05			; timer 1 counter high byte
T1L_l		=	$06			; timer 1 latch low byte
T1L_h		=	$07			; timer 1 latch high byte
T2C_l		=	$08			; timer 2 counter low byte
T2C_h		=	$09			; timer 2 counter high byte
SR		=	$0A			; shift register
ACR		=	$0B			; auxiliary control register
PCR		=	$0C			; peripheral control register
IFR		=	$0D			; interrupt flag register
IER		=	$0E			; interrupt enable register
ORAH		=	$0F			; data register

VIC_chip	=	$9000			; VIC chip base address
VIC_col	=	$9600			; colour RAM start
VIC_mem	=	$1E00			; video RAM start

IRQ_vec_l	=	$0314			; IRQ vector low byte
IRQ_vec_h	=	IRQ_vec_l+1		; IRQ vector high byte
BRK_vec_l	=	$0316			; BRK vector low byte
BRK_vec_h	=	BRK_vec_l+1		; BRK vector high byte
NMI_vec_l	=	$0318			; NMI vector low byte
NMI_vec_h	=	NMI_vec_l+1		; NMI vector high byte


; *********************************************************************
; periodic timer
; just does a simple moving thing to signal something is running

periodic
	LDA	cdtf_l			; get periodic timer f low byte
	BNE	LAB_209			; exit if not timed out

	LDA	#$C8				; about 200mS
	STA	cdtf_l			; reset periodic timer f low byte

	DEC	cdtf_h			; decrement periodic f timer byte
	BNE	LAB_209			; exit if not timed out

	LDA	#$02				; 2 x 200mS
	STA	cdtf_h			; reset periodic timer f high byte

						; just a scrolling star for now
	CLC					; clear carry for add
	LDA	logo_pos			; get logo position
	ADC	#$01				; increment it
	AND	#$03				; mask it
	STA	logo_pos			; save new value

	ADC	#$F5				; cursor low byte
	STA	cur_l				; save it
	LDA	#$1F				; cursor high byte
	STA	cur_h				; save it

	LDX	#$04				; 5 characters to do
logo_loop
	LDA	logo,X			; get byte
	JSR	chr_out			; out to screen
	DEX					; decrement count
	BPL	logo_loop			; loop if not all done

LAB_209
	RTS

logo
	.byte	"   * "			; scrolling logo bytes


; *********************************************************************
; initialise VIC 20 hardware

vicinit
	LDA	#$7F				; disable all interrupts
	STA	VIA_1+IER			; on VIA 1 ..
	STA	VIA_2+IER			; .. and VIA 2

	LDX	#$05				; set 6 vector bytes	
int_loop
	LDA	i_table,X			; get byte from vector table
	STA	IRQ_vec_l,X			; save to vector	
	DEX					; decrement count
	BPL	int_loop			; loop if more to do

	LDA	#$40				; timer 1 free run
	STA	VIA_2+ACR			; enable VIA 2 timer 1
	LDA	#$00				; handshake lines as inputs
	STA	VIA_1+ACR			; disable VIA 1 timer 1
	STA	VIA_1+PCR			; set VIA 1
	STA	VIA_2+PCR			; set VIA 2
	LDA	#$C0				; turn on timer interrupt
	STA	VIA_2+IER			; save to interrupt enable register
	LDA	#$E8				; counter low byte approx 1mS
	STA	VIA_2+T1C_l			; save to timer
	LDA	#$03				; counter high byte approx 1mS
	STA	VIA_2+T1C_h			; save to timer

	LDX	#$0F				; set 16 VIC registers	
vicchiploop
	LDA	vicset,X			; get byte from register set table
	STA	VIC_chip,X			; save to VIC chip	
	DEX					; decrement count
	BPL	vicchiploop			; loop if more to do

	CLI					; allow interrupts

	RTS


; *********************************************************************
; NMI interrupt routine - the NIC /should/ trigger this

iNMI
	SEI					; disable interrupts
	PHA					; save A

	LDA	NMI_res			; reset NMI latch
	LDA	com_reg_l			; read status byte
	AND	#$FE				; mask wanted bits
	ORA	IB				; OR with interrupt flag byte
	STA	IB				; save new interrupt flag byte

	PLA					; restore A
	RTI


;; *********************************************************************
;; IRQ interrupt routine - this part is in the kernal ROM and is here
;; only to show what happens before we get to our code.
;;
;;IRQ
;;	PHA					; save A
;;	TXA					; copy X
;;	PHA					; save X
;;	TYA					; copy Y
;;	PHA					; save Y
;;	TSX					; copy stack pointer
;;	LDA	stack+4,X			; get stacked status register	
;;	AND	#$10				; mask BRK flag
;;	BEQ	do_IRQ			; branch if not BRK
;;
;;	JMP	(BRK_vec_l)			; else do BRK vector (iBRK)
;;
;;do_IRQ					
;;	JMP	(IRQ_vec_l)			; do IRQ vector (iIRQ)
;;
;;
; *********************************************************************
; IRQ interrupt routine

iIRQ
	LDA	VIA_2+IFR			; check VIA 2 interrupt flags
	BPL	IRQexit			; exit if wasn't VIA 2

	AND	#$40				; mask timer 1 time-out flagged
	BEQ	IRQexit			; exit if not timer 1 time-out

	STA	VIA_2+IFR			; clear timer 1 time-out flag
	JSR	LAB_12000			; do timeout countdowns


; *********************************************************************
; BRK interrupt routine - does nothing at the moment

iBRK
IRQexit
	PLA					; pull byte
	TAY					; restore Y
	PLA					; pull byte
	TAX					; restore X
	PLA					; restore A
	RTI


; *********************************************************************
; setup screen & colour RAM, home cursor

setscr
	LDX	#$00				; clear index
	STX	cur_l				; clear cursor address low byte
	LDA	#$1E				; set high byte
	STA	cur_h				; set cursor address high byte

setscrloop
	LDA	vicscreen,X			; get character for low half of screen
	JSR	conv_ascii			; convert from ASCII
	STA	VIC_mem,X			; save to low half of video RAM
	LDA	vicscreen+$100,X		; get character for high half of screen
	JSR	conv_ascii			; convert from ASCII
	STA	VIC_mem+$100,X		; save to high half of video RAM
	LDA	#$06				; colour blue
	STA	VIC_col,X			; set low half of colour RAM
	STA	VIC_col+$100,X		; set high half of colour RAM
	DEX					; decrement count/index
	BNE	setscrloop			; loop if more to do

	RTS


; *********************************************************************
; convert ASCII to CBM chr set 2

conv_ascii
	CMP	#$61				; compare with "a"
	BCC	no_change			; branch if less

	CMP	#'z'+$01			; compare with max+1
	BCS	no_change			; branch if >=

	EOR	#$60				; convert to CBM character set
no_change
	RTS


; *********************************************************************
; brain dead (ish) screen print, prints chr in A

chr_out
	LDY	#$00				; clear index
	JSR	conv_ascii			; convert from ASCII
	STA	(cur_l),Y			; save to screen
	INC	cur_l				; increment low byte
	BNE	chr_exit			; exit if no rollover

	INC	cur_h				; increment high byte
chr_exit
	RTS


; *********************************************************************
; table of BCD values for each binary bit
; note - values are -1 as the add is always done with the carry set

b2b_table
	.byte	$63,$31,$15,$07,$03,$01,$00


; *********************************************************************
; default values for the interrupt vectors

i_table
	.word	iIRQ				; IRQ routine address
	.word	iBRK				; BRK routine address
	.word	iNMI				; NMI routine address


; *********************************************************************
; initial values for VIC registers

vicset	
;##	.byte	$05				; interlace & horizontal center [NTSC]
	.byte	$0C				; interlace & horizontal center [PAL]
						; bit	function
						; ---	--------
						;  7	interlace / non interlace
						; 6-0	horizontal origin
;##	.byte	$19				; vertical origin [NTSC]
	.byte	$26				; vertical origin [PAL]
	.byte	$96				; video address and colums ($9600 for colour RAM)
						; bit	function
						; ---	--------
						;  7	video address va9
						; 6-0	number of columns
	.byte	$2E				; rows & character size
						; bit	function
						; ---	--------
						;  7	b9 raster line
						; 6-1	number of rows
						;  0	8x16 / 8x8 characters
	.byte	$00				; raster line
	.byte	$F2				; video memory addresses ($1E00)
						; bit	function
						; ---	--------
						;  7	must be 1
						; 6-4	video memory address va12-va10
						; 3-0	character memory start address

						; 0000 ROM	$8000	set 1
						; 0001  "	$8400
						; 0010  "	$8800 set 2 - we use this
						; 0011  "	$8C00
						; 1100 RAM	$1000
						; 1101  "	$1400
						; 1110  "	$1800
						; 1111  "	$1C00

	.byte	$00				; light pen horizontal position
	.byte	$00				; light pen vertical position

	.byte	$00				; paddle X
	.byte	$00				; paddle Y
	.byte	$00				; oscillator 1 frequency
	.byte	$00				; oscillator 2 frequency
	.byte	$00				; oscillator 3 frequency
	.byte	$00				; noise source frequency
	.byte	$00				; aux colour & volume
						; bit	function
						; ---	--------
						; 7-4	auxiliary color information
						; 3-0	volume
	.byte	$1B				; screen and border colour
						; bit	function
						; ---	--------
						; 7-4	background colour
						;  3	inverted or normal mode
						; 2-0	border colour


; *********************************************************************
; some strings

; the initial screen for the server
; this is copied to the screen memory instead of clearing the screen
; it's big, it's ugly, but it works and is easy to change

; characters are all ASCII and converted to Vic characters on the fly
; I'm lazy, let the computer do the hard stuff

vicscreen
	.byte	"VIC 20 Mini web server"	; 1E00
	.byte	"                      "	; 1E16
	.byte	"   L.Davison 2003/4   "	; 1E2C
	.byte	"                      "	; 1E42
	.byte	"                      "	; 1E58
	.byte	"MAC:   x              "	; 1E6E
	.byte	"                      "	; 1E84
	.byte	"IP adr:x              "	; 1E9A
	.byte	"                      "	; 1EB0
	.byte	"TCP state: x          "	; 1EC6
	.byte	"                      "	; 1EDC
	.byte	"                      "	; 1EF2
	.byte	"                      "	; 1F08
	.byte	"                      "	; 1F1E
	.byte	"                      "	; 1F34
	.byte	"                      "	; 1F4A
	.byte	"                      "	; 1F60
	.byte	"                      "	; 1F76
	.byte	"                      "	; 1F8C
	.byte	"                      "	; 1FA2
	.byte	"                      "	; 1FB8
	.byte	"                      "	; 1FCE
	.byte	"00.44                 "	; 1FE4
;	.byte	"123456"				; 1FFA pad bytes

tcp_string
	.byte	"Listen    "
	.byte	"SYN recv'd"
	.byte	"Estab'sh'd"
	.byte	"Fin wait 1"
	.byte	"Fin wait 2"


; *********************************************************************
